"""
    Test Validation
    ~~~~~~~~~~~~~~~

    Tests for the Marshmallow extensions.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
from datetime import datetime
from importlib.metadata import version
from unittest import TestCase
from pytest import mark

from marshmallow import Schema
from marshmallow import ValidationError
from marshmallow import fields
from marshmallow.validate import OneOf, Range
from werkzeug.datastructures import MultiDict

from flask_rebar import compat
from flask_rebar import rebar
from flask_rebar.utils.request_utils import normalize_schema
from flask_rebar.validation import RequireOnDumpMixin
from flask_rebar.validation import CommaSeparatedList
from flask_rebar.validation import QueryParamList
from tests.test_rebar import create_rebar_app

MARSHMALLOW_VERSION = tuple(int(x) for x in version("marshmallow").split("."))


class NoRequireOnDumpMixinSchema(Schema):
    optional = fields.Str()
    value_optional = fields.Str(required=True, allow_none=True)
    value_required = fields.Str(required=True, allow_none=False)
    validation_required = fields.DateTime(required=True, allow_none=False)
    one_of_validation = fields.String(required=True, validate=OneOf(["a", "b"]))
    dump_only = fields.Integer(dump_only=True)  # type: ignore


class RequireOnDumpMixinSchema(NoRequireOnDumpMixinSchema, RequireOnDumpMixin):
    pass


class InnerNested(Schema, RequireOnDumpMixin):
    inner_dump_only = fields.String(dump_default="Inner dump-only", dump_only=True)
    inner_nested = fields.Nested(RequireOnDumpMixinSchema)
    inner_nested_dump_only = fields.Nested(RequireOnDumpMixinSchema, dump_only=True)
    inner_nested_list = fields.List(fields.Nested(RequireOnDumpMixinSchema))
    validated = fields.Integer(required=False, validate=Range(42, 99))


class OuterNested(Schema, RequireOnDumpMixin):
    outer_dump_only = fields.String(dump_default="Outer dump-only", dump_only=True)
    nested = fields.Nested(InnerNested)
    nested_list = fields.List(fields.Nested(InnerNested))


class OuterNestedNone(Schema, RequireOnDumpMixin):
    nested = fields.Nested(InnerNested, allow_none=True)


class RequireOutputMixinTest(TestCase):
    def setUp(self):
        super().setUp()
        self.validated_schema = normalize_schema(RequireOnDumpMixinSchema)
        self.unvalidated_schema = normalize_schema(NoRequireOnDumpMixinSchema)
        self.data = {
            "value_required": "abc",
            "value_optional": None,
            "validation_required": datetime.now(),
            "one_of_validation": "a",
        }

    def test_nominal(self):
        self.validated_schema.dump(self.data)

    def test_required_missing(self):
        del self.data["value_required"]
        with self.assertRaises(ValidationError) as ctx:
            compat.dump(self.validated_schema, self.data)
        self.assertIn("value_required", ctx.exception.messages)

    def test_required_none(self):
        self.data["value_required"] = None
        with self.assertRaises(ValidationError) as ctx:
            compat.dump(self.validated_schema, self.data)
        self.assertIn("value_required", ctx.exception.messages)

    def test_value_optional_missing(self):
        del self.data["value_optional"]
        with self.assertRaises(ValidationError) as ctx:
            compat.dump(self.validated_schema, self.data)
        self.assertIn("value_optional", ctx.exception.messages)

    def test_validation_works(self):
        self.data["validation_required"] = "123"
        with self.assertRaises(ValidationError) as ctx:
            compat.dump(self.validated_schema, self.data)
        # it's some sort of date error
        self.assertIn(
            "'str' object has no attribute 'isoformat'", ctx.exception.messages[0]
        )

    def test_required_failed_validate(self):
        self.data["one_of_validation"] = "c"
        with self.assertRaises(ValidationError) as ctx:
            compat.dump(self.validated_schema, self.data)
        self.assertIn("one_of_validation", ctx.exception.messages)

    def test_dump_only(self):
        self.data["dump_only"] = 42
        result = compat.dump(self.validated_schema, self.data)
        self.assertEqual(result["dump_only"], 42)

    def test_validation_opt_in(self):
        rebar_instance = rebar.Rebar()
        app = create_rebar_app(rebar_instance)

        self.data["one_of_validation"] = "c"
        with app.app_context():
            # explicit global opt-in:
            rebar_instance.validate_on_dump = True
            with self.assertRaises(ValidationError):
                compat.dump(self.unvalidated_schema, self.data)

            # explicit global opt-out:
            rebar_instance.validate_on_dump = False
            result = compat.dump(self.unvalidated_schema, self.data)
            self.assertEqual(result["one_of_validation"], "c")  # invalid but no error

            # with explicit opt-out, schema without mixin:
            result = compat.dump(self.unvalidated_schema, self.data)
            self.assertEqual(result["one_of_validation"], "c")  # invalid but no error

            # with explicit opt-out, but schema requiring val:
            with self.assertRaises(ValidationError):
                compat.dump(self.validated_schema, self.data)

    def test_top_level_list(self):
        data = [self.data.copy() for _ in range(3)]
        schema = RequireOnDumpMixinSchema(many=True)
        result = compat.dump(schema, data)
        self.assertEqual(len(result), len(data))
        data[1]["one_of_validation"] = "c"
        with self.assertRaises(ValidationError):
            compat.dump(schema, data)


class TestComplexNesting(TestCase):
    """
    Test cases for compat.dump with various cases in nested schemas.
    Note we use RequireOnDumpMixin in these Schemas so validation is always on.
    """

    def setUp(self):
        self.base_valid_inner_data = {
            "value_required": "abc",
            "value_optional": None,
            "validation_required": datetime.now(),
            "one_of_validation": "a",
        }

        self.base_valid_inner_wrapper = {
            "inner_nested": self.base_valid_inner_data.copy(),
            "inner_nested_list": [self.base_valid_inner_data.copy() for _ in range(3)],
        }

        self.base_valid_outer_wrapper = {
            "nested": self.base_valid_inner_wrapper.copy(),
            "nested_list": [self.base_valid_inner_wrapper.copy() for _ in range(3)],
        }

    @mark.skipif(
        MARSHMALLOW_VERSION < (3, 5),
        reason="https://github.com/marshmallow-code/marshmallow/issues/1497",
    )
    def test_dump_only_fields_defaults(self):
        """Our handling of dump_only fields respects defaults"""
        result = compat.dump(OuterNested(), self.base_valid_outer_wrapper)
        self.assertEqual(result["outer_dump_only"], "Outer dump-only")
        self.assertEqual(result["nested"]["inner_dump_only"], "Inner dump-only")
        self.assertNotIn(
            "dump_only", result["nested"]["inner_nested"]
        )  # no default for this one..
        for nested_item in result["nested_list"]:
            self.assertEqual(nested_item["inner_dump_only"], "Inner dump-only")
            for nested_nested_item in nested_item["inner_nested_list"]:
                self.assertNotIn(
                    "dump_only", nested_nested_item
                )  # no default for this one

    @mark.skipif(
        MARSHMALLOW_VERSION < (3, 5),
        reason="https://github.com/marshmallow-code/marshmallow/issues/1497",
    )
    def test_dump_only_fields_specified(self):
        """Our handling of dump_only fields preserves provided values"""
        self.base_valid_outer_wrapper["outer_dump_only"] = "Outer supplied"
        self.base_valid_outer_wrapper["nested"]["inner_dump_only"] = "Inner supplied"
        self.base_valid_outer_wrapper["nested"]["inner_nested"]["dump_only"] = 42
        self.base_valid_inner_wrapper["inner_nested_list"][1]["dump_only"] = 99
        self.base_valid_outer_wrapper["nested_list"][1][
            "inner_dump_only"
        ] = "Inner nested 1"
        result = compat.dump(OuterNested(), self.base_valid_outer_wrapper)
        self.assertEqual(result["outer_dump_only"], "Outer supplied")
        self.assertEqual(result["nested"]["inner_dump_only"], "Inner supplied")
        self.assertEqual(result["nested"]["inner_nested"]["dump_only"], 42)
        for i, nested_item in enumerate(result["nested_list"]):
            if i == 1:
                self.assertEqual(nested_item["inner_dump_only"], "Inner nested 1")
            else:
                self.assertEqual(nested_item["inner_dump_only"], "Inner dump-only")
            for j, nested_nested_item in enumerate(nested_item["inner_nested_list"]):
                if j == 1:
                    self.assertEqual(nested_nested_item["dump_only"], 99)
                else:
                    self.assertNotIn("dump_only", nested_nested_item)

    def test_validation_nested(self):
        """compat.dump validation works with nested schemas"""
        data = self.base_valid_outer_wrapper
        data["nested"]["validated"] = 999
        with self.assertRaises(ValidationError):
            compat.dump(OuterNested(), data)

    def test_validation_deep_nested(self):
        """compat.dump validation works with deeply nested schemas"""
        data = self.base_valid_outer_wrapper
        data["nested"]["inner_nested"]["one_of_validation"] = "z"
        with self.assertRaises(ValidationError):
            compat.dump(OuterNested(), data)

    def test_validation_nested_list(self):
        """compat.dump validation works with lists of nested schemas"""
        data = self.base_valid_outer_wrapper
        data["nested"]["inner_nested_list"][2]["one_of_validation"] = "z"
        with self.assertRaises(ValidationError):
            compat.dump(OuterNested(), data)

    def test_validation_nested_none(self):
        """compat.dump validation works with allow_none nested schemas"""
        data = self.base_valid_outer_wrapper
        data["nested"] = None
        result = compat.dump(OuterNestedNone(), data)
        self.assertEqual(result["nested"], None)


class StringList(Schema):
    foos = CommaSeparatedList(fields.String())


class IntegerList(Schema):
    foos = CommaSeparatedList(fields.Integer())


class TestCommaSeparatedList(TestCase):
    def test_deserialize(self):
        data = compat.load(StringList(), {"foos": "bar"})
        self.assertEqual(data["foos"], ["bar"])

        data = compat.load(StringList(), {"foos": "bar,baz"})
        self.assertEqual(data["foos"], ["bar", "baz"])

        data = compat.load(IntegerList(), {"foos": "1,2"})
        self.assertEqual(data["foos"], [1, 2])

    def test_serialize(self):
        data = compat.dump(StringList(), {"foos": ["bar"]})
        self.assertEqual(data["foos"], "bar")

        data = compat.dump(StringList(), {"foos": ["bar", "baz"]})
        self.assertEqual(data["foos"], "bar,baz")

        data = compat.dump(IntegerList(), {"foos": [1, 2]})
        self.assertEqual(data["foos"], "1,2")

    def test_deserialize_errors(self):
        with self.assertRaises(ValidationError) as ctx:
            compat.load(IntegerList(), {"foos": "1,two"})

        self.assertEqual(
            ctx.exception.messages, {"foos": {1: ["Not a valid integer."]}}
        )

    def test_serialize_errors(self):
        rebar_instance = rebar.Rebar()
        app = create_rebar_app(rebar_instance)
        with app.app_context():
            rebar_instance.validate_on_dump = (
                True  # cause ValueError to be wrapped in ValidationError
            )
            with self.assertRaises(ValidationError) as ctx:
                compat.dump(IntegerList(), {"foos": [42, "two"]})

            self.assertEqual(
                ctx.exception.messages,
                ["invalid literal for int() with base 10: 'two'"],
            )


class StringQuery(Schema):
    foos = QueryParamList(fields.String())


class IntegerQuery(Schema):
    foos = QueryParamList(fields.Integer())


class TestQueryParamList(TestCase):
    def test_deserialize(self):
        query = MultiDict([("foos", "bar")])
        data = compat.load(StringQuery(), query)
        self.assertEqual(data["foos"], ["bar"])

        query = MultiDict([("foos", "bar"), ("foos", "baz")])
        data = compat.load(StringQuery(), query)
        self.assertEqual(data["foos"], ["bar", "baz"])

        query = MultiDict([("foos", 1), ("foos", 2)])
        data = compat.load(IntegerQuery(), query)
        self.assertEqual(data["foos"], [1, 2])

    def test_deserialize_errors(self):
        query = MultiDict([("foos", 1), ("foos", "two")])

        with self.assertRaises(ValidationError) as ctx:
            compat.load(IntegerQuery(), query)

        self.assertEqual(
            ctx.exception.messages, {"foos": {1: ["Not a valid integer."]}}
        )


class DataKeySchema(RequireOnDumpMixin, Schema):
    test_field = fields.String(data_key="testField")


class TestDataKey(TestCase):
    def test_dump_and_validate_with_data_key(self):
        result = compat.dump(DataKeySchema(), {"test_field": "abc"})
        self.assertEqual(result, {"testField": "abc"})



"""
    Swagger V3 Generator
    ~~~~~~~~~~~~~~~~~~~~

    Class for converting a handler registry into a Swagger v3 specification.

    :copyright: Copyright 2019 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Sequence

from marshmallow import Schema

from flask_rebar.utils.defaults import USE_DEFAULT
from flask_rebar.swagger_generation import swagger_words as sw
from flask_rebar.swagger_generation.authenticator_to_swagger import (
    AuthenticatorConverterRegistry,
)
from flask_rebar.swagger_generation.swagger_generator_base import SwaggerGenerator
from flask_rebar.swagger_generation.swagger_objects import Server
from flask_rebar.swagger_generation.generator_utils import (
    format_path_for_swagger,
    verify_parameters_are_the_same,
    get_response_description,
    get_unique_schema_definitions,
    recursively_convert_dict_to_ordered_dict,
    get_ref_schema,
    get_unique_authenticators,
)
from flask_rebar.swagger_generation.marshmallow_to_swagger import (
    get_swagger_title,
    get_schema_fields,
    ConverterRegistry,
)
from flask_rebar.validation import Error

if TYPE_CHECKING:
    from flask_rebar import Tag
    from flask_rebar.rebar import HandlerRegistry, PathDefinition


class SwaggerV3Generator(SwaggerGenerator):
    """Generates a v3.1.0 Swagger specification from a Rebar object.

    Not all things are retrievable from the Rebar object, so this
    guy also needs some additional information to complete the job.

    :param ConverterRegistry query_string_converter_registry:
    :param ConverterRegistry request_body_converter_registry:
    :param ConverterRegistry headers_converter_registry:
    :param ConverterRegistry response_converter_registry:
        ConverterRegistrys that will be used to convert Marshmallow schemas
        to the corresponding types of swagger objects. These default to the
        global registries.

    :param Sequence[Tag] tags:
        A list of tags used by the specification with additional metadata.
    :param Sequence[Server] servers:
        A list of Server Objects to set as the server metadata for the specification.
    """

    _open_api_version = "3.1.0"

    def __init__(
        self,
        version: str = "1.0.0",
        title: str = "My API",
        description: str = "",
        query_string_converter_registry: Optional[ConverterRegistry] = None,
        request_body_converter_registry: Optional[ConverterRegistry] = None,
        headers_converter_registry: Optional[ConverterRegistry] = None,
        response_converter_registry: Optional[ConverterRegistry] = None,
        tags: Optional[Sequence["Tag"]] = None,
        servers: Optional[List[Server]] = None,
        default_response_schema: Schema = Error(),
        authenticator_converter_registry: Optional[
            AuthenticatorConverterRegistry
        ] = None,
        include_hidden: bool = False,
    ):
        super().__init__(
            openapi_major_version=3,
            version=version,
            title=title,
            description=description,
            default_response_schema=default_response_schema,
            query_string_converter_registry=query_string_converter_registry,
            request_body_converter_registry=request_body_converter_registry,
            headers_converter_registry=headers_converter_registry,
            response_converter_registry=response_converter_registry,
            authenticator_converter_registry=authenticator_converter_registry,
            include_hidden=include_hidden,
        )
        self.tags = tags
        self.servers = servers
        self._ref_base = "#/components/schemas"

    def generate_swagger(
        self, registry: "HandlerRegistry", host: Optional[str] = None
    ) -> Dict[str, Any]:
        return self.generate(registry=registry, host=host)

    def generate(
        self,
        registry: "HandlerRegistry",
        host: Optional[str] = None,
        sort_keys: bool = True,
    ) -> Dict[str, Any]:
        """Generates a Swagger specification from a Rebar instance.

        :param flask_rebar.rebar.HandlerRegistry registry:
        :param str host: Adds this host as a Server Object for the service
        :param bool sort_keys: Use OrderedDicts sorted by keys instead of dicts
        :rtype: dict
        """

        components = self._get_components(registry=registry)

        default_security = []
        for authenticator in registry.default_authenticators:
            default_security.extend(
                self.authenticator_converter.get_security_requirements(authenticator)
            )

        paths = self._get_paths(
            paths=registry.paths,
            default_headers_schema=registry.default_headers_schema,
            default_security=default_security,
        )

        swagger: Dict[str, Any] = {
            sw.openapi: self.get_open_api_version(),
            sw.info: self._get_info(),
            sw.paths: paths,
            sw.components: components,
        }
        if default_security:
            swagger[sw.security] = default_security

        if self.tags:
            swagger[sw.tags] = [tag.as_swagger() for tag in self.tags]

        servers = self.servers or []

        if host:
            servers.append(Server(url=host))

        if servers:
            swagger[sw.servers] = [server.as_swagger() for server in servers]

        if sort_keys:
            # Sort the swagger we generated by keys to produce a consistent output.
            swagger = recursively_convert_dict_to_ordered_dict(swagger)

        return swagger

    def _get_paths(
        self,
        paths: Dict[str, Dict[str, "PathDefinition"]],
        default_headers_schema: Optional[Schema],
        default_security: Optional[Any] = None,
    ) -> Dict[str, Any]:
        path_definitions: Dict[str, Any] = {}

        for path, methods in paths.items():
            spec_path, path_args = format_path_for_swagger(path)

            if not self.include_hidden and all(
                d.hidden for method, d in methods.items()
            ):
                continue

            # Different Flask paths might correspond to the same Swagger path
            # because of Flask URL path converters. In this case, let's just
            # work off the same path definitions.
            if spec_path in path_definitions:
                path_definition = path_definitions[spec_path]
            else:
                path_definitions[spec_path] = path_definition = {}

            if path_args:
                path_params = []
                for path_arg in path_args:
                    next_param = {
                        sw.name: path_arg.name,
                        sw.required: True,
                        sw.in_: sw.path,
                        sw.style: sw.simple,
                        sw.schema: {},
                    }
                    if hasattr(
                        converter := self.flask_converters_to_swagger_types[
                            path_arg.type
                        ],
                        "to_swagger",
                    ):
                        next_param[sw.schema].update(converter.to_swagger())
                    else:
                        next_param[sw.schema][sw.type_] = converter
                    path_params.append(next_param)

                # We have to check for an ugly case here. If different Flask
                # paths that map to the same Swagger path use different URL
                # converters for the same parameter, we have a problem. Let's
                # just throw an error in this case.
                if sw.parameters in path_definition:
                    verify_parameters_are_the_same(
                        path_definition[sw.parameters], path_params
                    )

                path_definition[sw.parameters] = path_params

            for method, d in methods.items():
                if not self.include_hidden and d.hidden:
                    continue

                responses_definition = {
                    sw.default: self._get_response_definition(
                        self.default_response_schema
                    )
                }

                if d.response_body_schema:
                    for status_code, schema in d.response_body_schema.items():
                        if schema is not None:
                            response_definition = self._get_response_definition(schema)
                        else:
                            response_definition = {sw.description: "No response body."}

                        responses_definition[str(status_code)] = response_definition

                parameters_definition = []

                if d.query_string_schema:
                    parameters_definition.extend(
                        self._convert_schema_to_list_of_parameters(
                            schema=d.query_string_schema,
                            converter=self._query_string_converter,
                            in_=sw.query,
                        )
                    )

                if d.headers_schema is USE_DEFAULT:
                    headers_schema = default_headers_schema
                else:
                    headers_schema = d.headers_schema

                if headers_schema:
                    parameters_definition.extend(
                        self._convert_schema_to_list_of_parameters(
                            schema=headers_schema,
                            converter=self._headers_converter,
                            in_=sw.header,
                        )
                    )

                request_body = None

                if d.request_body_schema:
                    schema = d.request_body_schema

                    request_body = {
                        sw.required: True,
                        sw.content: {
                            "application/json": {
                                sw.schema: get_ref_schema(self._ref_base, schema)
                            }
                        },
                    }

                method_lower = method.lower()
                path_definition[method_lower] = {
                    sw.operation_id: d.endpoint or get_swagger_title(d.func),
                    sw.responses: responses_definition,
                }

                if d.func.__doc__:
                    path_definition[method_lower][sw.description] = d.func.__doc__

                if parameters_definition:
                    path_definition[method_lower][sw.parameters] = parameters_definition

                if request_body:
                    path_definition[method_lower][sw.request_body] = request_body

                if not d.authenticators:
                    path_definition[method_lower][sw.security] = []
                else:
                    non_default = False
                    security = []
                    for authenticator in d.authenticators:
                        if authenticator is not USE_DEFAULT:
                            security.extend(
                                self.authenticator_converter.get_security_requirements(
                                    authenticator
                                )
                            )
                            non_default = True
                        elif default_security is not None:
                            security.extend(default_security)
                    if non_default:
                        path_definition[method_lower][sw.security] = security

                if d.tags:
                    path_definition[method_lower][sw.tags] = d.tags

                if d.summary:
                    path_definition[method_lower][sw.summary] = d.summary

        return path_definitions

    def _get_response_definition(self, schema: Schema) -> Dict[str, Any]:
        return {
            sw.description: get_response_description(schema),
            sw.content: {
                "application/json": {sw.schema: get_ref_schema(self._ref_base, schema)}
            },
        }

    def _get_components(self, registry: "HandlerRegistry") -> Dict[str, Any]:
        """

        :param flask_rebar.rebar.HandlerRegistry registry:
        :return:
        """
        components = {}

        schemas = get_unique_schema_definitions(
            registry=registry,
            base=self._ref_base,
            default_response_schema=self.default_response_schema,
            response_converter=self._response_converter,
            request_body_converter=self._request_body_converter,
        )
        if schemas:
            components[sw.schemas] = schemas

        security_schemes = {}
        authenticators = get_unique_authenticators(registry)
        for authenticator in authenticators:
            # We should probably eventually check that scheme with the same name are identical
            # rather than just overwriting the existing scheme definition.
            security_schemes.update(
                self.authenticator_converter.get_security_schemes(authenticator)
            )
        if security_schemes:
            components[sw.security_schemes] = security_schemes

        return components

    def _convert_schema_to_list_of_parameters(
        self, schema: Schema, converter: Callable, in_: str
    ) -> List[Dict[str, Any]]:
        """Swagger is only _based_ on JSONSchema. Query string and header parameters
        are represented as list, not as an object. This converts a JSONSchema
        object (as return by the converters) to a list of parameters suitable for
        swagger.

        :param marshmallow.Schema schema:
        :param str in_: 'query' or 'header'
        :rtype: list[dict]
        """
        parameters = []

        for prop, field in get_schema_fields(schema):
            jsonschema = converter(field)

            # Pardon the ugliness.
            # We need the "explode" key to be at the parameters level, not at the schema level.
            explode = jsonschema.pop(sw.explode, None)
            description = jsonschema.pop(sw.description, None)
            style = jsonschema.pop(sw.style, None)

            parameter = {
                sw.name: prop,
                sw.in_: in_,
                sw.schema: jsonschema,
                sw.required: bool(field.required),
            }

            if explode is not None:
                parameter[sw.explode] = explode
            if description is not None:
                parameter[sw.description] = description
            if style is not None:
                parameter[sw.style] = style

            parameters.append(parameter)

        return parameters



"""
    Header API Key Authenticator
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Authenticator for API key passed in header.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
from typing import Dict

from flask import request, g
from hmac import compare_digest

from flask_rebar import errors, messages
from flask_rebar.authenticators.base import Authenticator


def get_authenticated_app_name() -> str:
    return g.authenticated_app_name


class HeaderApiKeyAuthenticator(Authenticator):
    """
    Authenticates based on a small set of shared secrets, passed via a header.

    This allows multiple client applications to be registered with their own
    keys.
    This also allows multiple keys to be registered for a single client
    application.

    :param str header:
        The header where clients where client applications must include
        their secret.
    :param str name:
        A name for this authenticator. This should be unique across
        authenticators.
    """

    # This authenticator allows multiple applications to have different keys.
    # This is the default name, if someone doesn't need about this feature.
    DEFAULT_APP_NAME = "default"

    def __init__(self, header: str, name: str = "sharedSecret") -> None:
        self.header = header
        self.keys: Dict[str, str] = {}
        self.name = name

    @property
    def authenticated_app_name(self) -> str:
        return get_authenticated_app_name()

    def register_key(self, key: str, app_name: str = DEFAULT_APP_NAME) -> None:
        """
        Register a client application's shared secret.

        :param str app_name:
            Name for the application. Since an application can have multiple
            shared secrets, this does not need to be unique.
        :param str key:
            The shared secret.
        """
        self.keys[key] = app_name

    def authenticate(self) -> None:
        if self.header not in request.headers:
            raise errors.Unauthorized(messages.missing_auth_token)

        token = request.headers[self.header]

        for key, app_name in self.keys.items():
            if compare_digest(str(token), key):
                g.authenticated_app_name = app_name
                break
        else:
            raise errors.Unauthorized(messages.invalid_auth_token)



from enum import Enum

from todo.database import todo_id_sequence, todo_database
from todo.app import rebar
from todo.app import registry
from todo.schemas import (
    CreateTodoSchema,
    GetTodoListSchema,
    TodoResourceSchema,
    TodoListSchema,
    UpdateTodoSchema,
)


@registry.handles(
    rule="/todos",
    method="POST",
    request_body_schema=CreateTodoSchema(),
    tags=["todo"],
    # This dictionary tells framer which schema to use for which response code.
    # This is a little ugly, but tremendously helpful for generating swagger.
    response_body_schema={
        201: TodoResourceSchema()
    },  # for versions <= 1.7.0, use marshal_schema
)
def create_todo():
    global todo_id_sequence, todo_database

    # The body is eagerly validated with the `request_body_schema` provided in
    # the decorator. The resulting parameters are now available here:
    todo = rebar.validated_body

    todo_id_sequence += 1

    todo["id"] = todo_id_sequence
    todo_database[todo_id_sequence] = todo

    # The return value may be an object to encoded as JSON or a tuple where
    # the first item is the value to be encoded as JSON and the second is
    # the HTTP response code. In the case where no response code is included,
    # 200 is assumed.
    return todo, 201


@registry.handles(
    rule="/todos",
    method="GET",
    query_string_schema=GetTodoListSchema(),
    tags=["todo"],
    # If the value for this is not a dictionary, the response code is assumed
    # to be 200
    response_body_schema=TodoListSchema(),  # for versions <= 1.7.0, use marshal_schema
)
def get_todos():
    global todo_database

    # Just like validated_body, query string parameters are eagerly validated
    # and made available here. Flask-toolbox does treats a request body and
    # query string parameters as two separate sources, and currently does not
    # implement any abstraction on top of them.
    args = rebar.validated_args

    todos = todo_database.values()

    if "complete" in args:
        todos = [t for t in todos if t["complete"] == args["complete"]]

    return todos


@registry.handles(
    rule="/todos/<todo_types:todo_type>",
    method="GET",
    query_string_schema=GetTodoListSchema(),
    tags=["todo"],
    # If the value for this is not a dictionary, the response code is assumed
    # to be 200
    response_body_schema=TodoListSchema(),  # for versions <= 1.7.0, use marshal_schema
)
def get_todos_by_type(todo_type):
    global todo_database

    # Just like validated_body, query string parameters are eagerly validated
    # and made available here. Flask-toolbox does treats a request body and
    # query string parameters as two separate sources, and currently does not
    # implement any abstraction on top of them.
    args = rebar.validated_args

    todos = todo_database.values()

    if "complete" in args:
        todos = [t for t in todos if t["complete"] == args["complete"]]
    todos = [t for t in todos if t["type"] == todo_type.name]

    return todos


@registry.handles(
    rule="/todos/<int:todo_id>",
    method="PATCH",
    response_body_schema=TodoResourceSchema(),  # for versions <= 1.7.0, use marshal_schema
    request_body_schema=UpdateTodoSchema(),
    tags=["todo"],
)
def update_todo(todo_id):
    global todo_database

    if todo_id not in todo_database:
        raise errors.NotFound()

    params = rebar.validated_body
    todo_database[todo_id].update(params)
    todo = todo_database[todo_id]

    return todo



"""
    Messages
    ~~~~~~~~

    Helpers for generating messages that the API returns.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
from collections import namedtuple


# machine-friendly equivalents of associated human-friendly messages
class ErrorCode:
    BODY_VALIDATION_FAILED = "body_validation_failed"
    EMPTY_JSON_BODY = "empty_json_body"
    INTERNAL_SERVER_ERROR = "internal_server_error"
    INVALID_AUTH_TOKEN = "invalid_auth_token"
    INVALID_JSON = "invalid_json"
    MISSING_AUTH_TOKEN = "missing_auth_token"
    QUERY_STRING_VALIDATION_FAILED = "query_string_validation_failed"
    UNSUPPORTED_CONTENT_TYPE = "unsupported_content_type"
    HEADER_VALIDATION_FAILED = "header_validation_failed"
    REQUIRED_FIELD_MISSING = "required_fields_missing"
    REQUIRED_FIELD_EMPTY = "required_field_empty"
    UNSUPPORTED_FIELDS = "unsupported_fields"


ErrorMessage = namedtuple("ErrorMessage", "message, rebar_error_code")


body_validation_failed = ErrorMessage(
    "JSON body parameters are invalid.", ErrorCode.BODY_VALIDATION_FAILED
)

empty_json_body = ErrorMessage(
    "Fields must be in JSON body.", ErrorCode.EMPTY_JSON_BODY
)

internal_server_error = ErrorMessage(
    "Sorry, there was an internal error.", ErrorCode.INTERNAL_SERVER_ERROR
)

invalid_auth_token = ErrorMessage(
    "Invalid authentication.", ErrorCode.INVALID_AUTH_TOKEN
)

invalid_json = ErrorMessage("Failed to decode JSON body.", ErrorCode.INVALID_JSON)

missing_auth_token = ErrorMessage(
    "No auth token provided.", ErrorCode.MISSING_AUTH_TOKEN
)

query_string_validation_failed = ErrorMessage(
    "Query string parameters are invalid.", ErrorCode.QUERY_STRING_VALIDATION_FAILED
)

unsupported_content_type = ErrorMessage(
    "Only payloads with 'content-type' 'application/json' are supported.",
    ErrorCode.UNSUPPORTED_CONTENT_TYPE,
)

header_validation_failed = ErrorMessage(
    "Header parameters are invalid", ErrorCode.HEADER_VALIDATION_FAILED
)


def required_field_missing(field_name: str) -> ErrorMessage:
    return ErrorMessage(
        f"Required field missing: {field_name}",
        ErrorCode.REQUIRED_FIELD_MISSING,
    )


def required_field_empty(field_name: str) -> ErrorMessage:
    return ErrorMessage(
        f"Value for required field cannot be None: {field_name}",
        ErrorCode.REQUIRED_FIELD_EMPTY,
    )


def unsupported_fields(field_names: str) -> ErrorMessage:
    return ErrorMessage(
        "Unexpected field: {}".format(",".join(field_names)),
        ErrorCode.UNSUPPORTED_FIELDS,
    )



# This is a JSONSchema document that validates a Swagger spec
# We can use this to quickly check if the swagger we are producing is valid

LONG_DESCRIPTION = (
    "A longer description - should be different from the title. "
    + "GitHub Flavored Markdown is allowed."
)

SWAGGER_V2_JSONSCHEMA = {
    "title": "A JSON Schema for Swagger 2.0 API.",
    "id": "http://swagger.io/v2/schema.json#",
    "$schema": "http://json-schema.org/draft-04/schema#",
    "type": "object",
    "required": ["swagger", "info", "paths"],
    "additionalProperties": False,
    "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
    "properties": {
        "swagger": {
            "type": "string",
            "enum": ["2.0"],
            "description": "The Swagger version of this document.",
        },
        "info": {"$ref": "#/definitions/info"},
        "host": {
            "type": "string",
            "pattern": "^[^{}/ :\\\\]+(?::\\d+)?$",
            "description": "The host (name or ip) of the API. Example: 'swagger.io'",
        },
        "basePath": {
            "type": "string",
            "pattern": "^/",
            "description": "The base path to the API. Example: '/api'.",
        },
        "schemes": {"$ref": "#/definitions/schemesList"},
        "consumes": {
            "description": "A list of MIME types accepted by the API.",
            "allOf": [{"$ref": "#/definitions/mediaTypeList"}],
        },
        "produces": {
            "description": "A list of MIME types the API can produce.",
            "allOf": [{"$ref": "#/definitions/mediaTypeList"}],
        },
        "paths": {"$ref": "#/definitions/paths"},
        "definitions": {"$ref": "#/definitions/definitions"},
        "parameters": {"$ref": "#/definitions/parameterDefinitions"},
        "responses": {"$ref": "#/definitions/responseDefinitions"},
        "security": {"$ref": "#/definitions/security"},
        "securityDefinitions": {"$ref": "#/definitions/securityDefinitions"},
        "tags": {
            "type": "array",
            "items": {"$ref": "#/definitions/tag"},
            "uniqueItems": True,
        },
        "externalDocs": {"$ref": "#/definitions/externalDocs"},
    },
    "definitions": {
        "info": {
            "type": "object",
            "description": "General information about the API.",
            "required": ["version", "title"],
            "additionalProperties": False,
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
            "properties": {
                "title": {
                    "type": "string",
                    "description": "A unique and precise title of the API.",
                },
                "version": {
                    "type": "string",
                    "description": "A semantic version number of the API.",
                },
                "description": {
                    "type": "string",
                    "description": LONG_DESCRIPTION,
                },
                "termsOfService": {
                    "type": "string",
                    "description": "The terms of service for the API.",
                },
                "contact": {"$ref": "#/definitions/contact"},
                "license": {"$ref": "#/definitions/license"},
            },
        },
        "contact": {
            "type": "object",
            "description": "Contact information for the owners of the API.",
            "additionalProperties": False,
            "properties": {
                "name": {
                    "type": "string",
                    "description": "The identifying name of the contact person/organization.",
                },
                "url": {
                    "type": "string",
                    "description": "The URL pointing to the contact information.",
                    "format": "uri",
                },
                "email": {
                    "type": "string",
                    "description": "The email address of the contact person/organization.",
                    "format": "email",
                },
            },
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
        },
        "license": {
            "type": "object",
            "required": ["name"],
            "additionalProperties": False,
            "properties": {
                "name": {
                    "type": "string",
                    "description": "The name of the license type. It's encouraged to use an OSI compatible license.",
                },
                "url": {
                    "type": "string",
                    "description": "The URL pointing to the license.",
                    "format": "uri",
                },
            },
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
        },
        "paths": {
            "type": "object",
            "description": "Relative paths to the individual endpoints. They must be relative to the 'basePath'.",
            "patternProperties": {
                "^x-": {"$ref": "#/definitions/vendorExtension"},
                "^/": {"$ref": "#/definitions/pathItem"},
            },
            "additionalProperties": False,
        },
        "definitions": {
            "type": "object",
            "additionalProperties": {"$ref": "#/definitions/schema"},
            "description": "One or more JSON objects describing the schemas being consumed and produced by the API.",
        },
        "parameterDefinitions": {
            "type": "object",
            "additionalProperties": {"$ref": "#/definitions/parameter"},
            "description": "One or more JSON representations for parameters",
        },
        "responseDefinitions": {
            "type": "object",
            "additionalProperties": {"$ref": "#/definitions/response"},
            "description": "One or more JSON representations for responses",
        },
        "externalDocs": {
            "type": "object",
            "additionalProperties": False,
            "description": "information about external documentation",
            "required": ["url"],
            "properties": {
                "description": {"type": "string"},
                "url": {"type": "string", "format": "uri"},
            },
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
        },
        "examples": {"type": "object", "additionalProperties": True},
        "mimeType": {
            "type": "string",
            "description": "The MIME type of the HTTP message.",
        },
        "operation": {
            "type": "object",
            "required": ["responses"],
            "additionalProperties": False,
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
            "properties": {
                "tags": {
                    "type": "array",
                    "items": {"type": "string"},
                    "uniqueItems": True,
                },
                "summary": {
                    "type": "string",
                    "description": "A brief summary of the operation.",
                },
                "description": {
                    "type": "string",
                    "description": "A longer description of the operation, GitHub Flavored Markdown is allowed.",
                },
                "externalDocs": {"$ref": "#/definitions/externalDocs"},
                "operationId": {
                    "type": "string",
                    "description": "A unique identifier of the operation.",
                },
                "produces": {
                    "description": "A list of MIME types the API can produce.",
                    "allOf": [{"$ref": "#/definitions/mediaTypeList"}],
                },
                "consumes": {
                    "description": "A list of MIME types the API can consume.",
                    "allOf": [{"$ref": "#/definitions/mediaTypeList"}],
                },
                "parameters": {"$ref": "#/definitions/parametersList"},
                "responses": {"$ref": "#/definitions/responses"},
                "schemes": {"$ref": "#/definitions/schemesList"},
                "deprecated": {"type": "boolean", "default": False},
                "security": {"$ref": "#/definitions/security"},
            },
        },
        "pathItem": {
            "type": "object",
            "additionalProperties": False,
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
            "properties": {
                "$ref": {"type": "string"},
                "get": {"$ref": "#/definitions/operation"},
                "put": {"$ref": "#/definitions/operation"},
                "post": {"$ref": "#/definitions/operation"},
                "delete": {"$ref": "#/definitions/operation"},
                "options": {"$ref": "#/definitions/operation"},
                "head": {"$ref": "#/definitions/operation"},
                "patch": {"$ref": "#/definitions/operation"},
                "parameters": {"$ref": "#/definitions/parametersList"},
            },
        },
        "responses": {
            "type": "object",
            "description": "Response objects names can either be any valid HTTP status code or 'default'.",
            "minProperties": 1,
            "additionalProperties": False,
            "patternProperties": {
                "^([0-9]{3})$|^(default)$": {"$ref": "#/definitions/responseValue"},
                "^x-": {"$ref": "#/definitions/vendorExtension"},
            },
            "not": {
                "type": "object",
                "additionalProperties": False,
                "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
            },
        },
        "responseValue": {
            "oneOf": [
                {"$ref": "#/definitions/response"},
                {"$ref": "#/definitions/jsonReference"},
            ]
        },
        "response": {
            "type": "object",
            "required": ["description"],
            "properties": {
                "description": {"type": "string"},
                "schema": {
                    "oneOf": [
                        {"$ref": "#/definitions/schema"},
                        {"$ref": "#/definitions/fileSchema"},
                    ]
                },
                "headers": {"$ref": "#/definitions/headers"},
                "examples": {"$ref": "#/definitions/examples"},
            },
            "additionalProperties": False,
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
        },
        "headers": {
            "type": "object",
            "additionalProperties": {"$ref": "#/definitions/header"},
        },
        "header": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["string", "number", "integer", "boolean", "array"],
                },
                "format": {"type": "string"},
                "items": {"$ref": "#/definitions/primitivesItems"},
                "collectionFormat": {"$ref": "#/definitions/collectionFormat"},
                "default": {"$ref": "#/definitions/default"},
                "maximum": {"$ref": "#/definitions/maximum"},
                "exclusiveMaximum": {"$ref": "#/definitions/exclusiveMaximum"},
                "minimum": {"$ref": "#/definitions/minimum"},
                "exclusiveMinimum": {"$ref": "#/definitions/exclusiveMinimum"},
                "maxLength": {"$ref": "#/definitions/maxLength"},
                "minLength": {"$ref": "#/definitions/minLength"},
                "pattern": {"$ref": "#/definitions/pattern"},
                "maxItems": {"$ref": "#/definitions/maxItems"},
                "minItems": {"$ref": "#/definitions/minItems"},
                "uniqueItems": {"$ref": "#/definitions/uniqueItems"},
                "enum": {"$ref": "#/definitions/enum"},
                "multipleOf": {"$ref": "#/definitions/multipleOf"},
                "description": {"type": "string"},
            },
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
        },
        "vendorExtension": {
            "description": "Any property starting with x- is valid.",
            "additionalProperties": True,
            "additionalItems": True,
        },
        "bodyParameter": {
            "type": "object",
            "required": ["name", "in", "schema"],
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
            "properties": {
                "description": {
                    "type": "string",
                    "description": LONG_DESCRIPTION,
                },
                "name": {"type": "string", "description": "The name of the parameter."},
                "in": {
                    "type": "string",
                    "description": "Determines the location of the parameter.",
                    "enum": ["body"],
                },
                "required": {
                    "type": "boolean",
                    "description": "Determines whether or not this parameter is required or optional.",
                    "default": False,
                },
                "schema": {"$ref": "#/definitions/schema"},
            },
            "additionalProperties": False,
        },
        "headerParameterSubSchema": {
            "additionalProperties": False,
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
            "properties": {
                "required": {
                    "type": "boolean",
                    "description": "Determines whether or not this parameter is required or optional.",
                    "default": False,
                },
                "in": {
                    "type": "string",
                    "description": "Determines the location of the parameter.",
                    "enum": ["header"],
                },
                "description": {
                    "type": "string",
                    "description": LONG_DESCRIPTION,
                },
                "name": {"type": "string", "description": "The name of the parameter."},
                "type": {
                    "type": "string",
                    "enum": ["string", "number", "boolean", "integer", "array"],
                },
                "format": {"type": "string"},
                "items": {"$ref": "#/definitions/primitivesItems"},
                "collectionFormat": {"$ref": "#/definitions/collectionFormat"},
                "default": {"$ref": "#/definitions/default"},
                "maximum": {"$ref": "#/definitions/maximum"},
                "exclusiveMaximum": {"$ref": "#/definitions/exclusiveMaximum"},
                "minimum": {"$ref": "#/definitions/minimum"},
                "exclusiveMinimum": {"$ref": "#/definitions/exclusiveMinimum"},
                "maxLength": {"$ref": "#/definitions/maxLength"},
                "minLength": {"$ref": "#/definitions/minLength"},
                "pattern": {"$ref": "#/definitions/pattern"},
                "maxItems": {"$ref": "#/definitions/maxItems"},
                "minItems": {"$ref": "#/definitions/minItems"},
                "uniqueItems": {"$ref": "#/definitions/uniqueItems"},
                "enum": {"$ref": "#/definitions/enum"},
                "multipleOf": {"$ref": "#/definitions/multipleOf"},
            },
        },
        "queryParameterSubSchema": {
            "additionalProperties": False,
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
            "properties": {
                "required": {
                    "type": "boolean",
                    "description": "Determines whether or not this parameter is required or optional.",
                    "default": False,
                },
                "in": {
                    "type": "string",
                    "description": "Determines the location of the parameter.",
                    "enum": ["query"],
                },
                "description": {
                    "type": "string",
                    "description": LONG_DESCRIPTION,
                },
                "name": {"type": "string", "description": "The name of the parameter."},
                "allowEmptyValue": {
                    "type": "boolean",
                    "default": False,
                    "description": "allows sending a parameter by name only or with an empty value.",
                },
                "type": {
                    "type": "string",
                    "enum": ["string", "number", "boolean", "integer", "array"],
                },
                "format": {"type": "string"},
                "items": {"$ref": "#/definitions/primitivesItems"},
                "collectionFormat": {"$ref": "#/definitions/collectionFormatWithMulti"},
                "default": {"$ref": "#/definitions/default"},
                "maximum": {"$ref": "#/definitions/maximum"},
                "exclusiveMaximum": {"$ref": "#/definitions/exclusiveMaximum"},
                "minimum": {"$ref": "#/definitions/minimum"},
                "exclusiveMinimum": {"$ref": "#/definitions/exclusiveMinimum"},
                "maxLength": {"$ref": "#/definitions/maxLength"},
                "minLength": {"$ref": "#/definitions/minLength"},
                "pattern": {"$ref": "#/definitions/pattern"},
                "maxItems": {"$ref": "#/definitions/maxItems"},
                "minItems": {"$ref": "#/definitions/minItems"},
                "uniqueItems": {"$ref": "#/definitions/uniqueItems"},
                "enum": {"$ref": "#/definitions/enum"},
                "multipleOf": {"$ref": "#/definitions/multipleOf"},
            },
        },
        "formDataParameterSubSchema": {
            "additionalProperties": False,
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
            "properties": {
                "required": {
                    "type": "boolean",
                    "description": "Determines whether or not this parameter is required or optional.",
                    "default": False,
                },
                "in": {
                    "type": "string",
                    "description": "Determines the location of the parameter.",
                    "enum": ["formData"],
                },
                "description": {
                    "type": "string",
                    "description": LONG_DESCRIPTION,
                },
                "name": {"type": "string", "description": "The name of the parameter."},
                "allowEmptyValue": {
                    "type": "boolean",
                    "default": False,
                    "description": "allows sending a parameter by name only or with an empty value.",
                },
                "type": {
                    "type": "string",
                    "enum": ["string", "number", "boolean", "integer", "array", "file"],
                },
                "format": {"type": "string"},
                "items": {"$ref": "#/definitions/primitivesItems"},
                "collectionFormat": {"$ref": "#/definitions/collectionFormatWithMulti"},
                "default": {"$ref": "#/definitions/default"},
                "maximum": {"$ref": "#/definitions/maximum"},
                "exclusiveMaximum": {"$ref": "#/definitions/exclusiveMaximum"},
                "minimum": {"$ref": "#/definitions/minimum"},
                "exclusiveMinimum": {"$ref": "#/definitions/exclusiveMinimum"},
                "maxLength": {"$ref": "#/definitions/maxLength"},
                "minLength": {"$ref": "#/definitions/minLength"},
                "pattern": {"$ref": "#/definitions/pattern"},
                "maxItems": {"$ref": "#/definitions/maxItems"},
                "minItems": {"$ref": "#/definitions/minItems"},
                "uniqueItems": {"$ref": "#/definitions/uniqueItems"},
                "enum": {"$ref": "#/definitions/enum"},
                "multipleOf": {"$ref": "#/definitions/multipleOf"},
            },
        },
        "pathParameterSubSchema": {
            "additionalProperties": False,
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
            "required": ["required"],
            "properties": {
                "required": {
                    "type": "boolean",
                    "enum": [True],
                    "description": "Determines whether or not this parameter is required or optional.",
                },
                "in": {
                    "type": "string",
                    "description": "Determines the location of the parameter.",
                    "enum": ["path"],
                },
                "description": {
                    "type": "string",
                    "description": LONG_DESCRIPTION,
                },
                "name": {"type": "string", "description": "The name of the parameter."},
                "type": {
                    "type": "string",
                    "enum": ["string", "number", "boolean", "integer", "array"],
                },
                "format": {"type": "string"},
                "items": {"$ref": "#/definitions/primitivesItems"},
                "collectionFormat": {"$ref": "#/definitions/collectionFormat"},
                "default": {"$ref": "#/definitions/default"},
                "maximum": {"$ref": "#/definitions/maximum"},
                "exclusiveMaximum": {"$ref": "#/definitions/exclusiveMaximum"},
                "minimum": {"$ref": "#/definitions/minimum"},
                "exclusiveMinimum": {"$ref": "#/definitions/exclusiveMinimum"},
                "maxLength": {"$ref": "#/definitions/maxLength"},
                "minLength": {"$ref": "#/definitions/minLength"},
                "pattern": {"$ref": "#/definitions/pattern"},
                "maxItems": {"$ref": "#/definitions/maxItems"},
                "minItems": {"$ref": "#/definitions/minItems"},
                "uniqueItems": {"$ref": "#/definitions/uniqueItems"},
                "enum": {"$ref": "#/definitions/enum"},
                "multipleOf": {"$ref": "#/definitions/multipleOf"},
            },
        },
        "nonBodyParameter": {
            "type": "object",
            "required": ["name", "in", "type"],
            "oneOf": [
                {"$ref": "#/definitions/headerParameterSubSchema"},
                {"$ref": "#/definitions/formDataParameterSubSchema"},
                {"$ref": "#/definitions/queryParameterSubSchema"},
                {"$ref": "#/definitions/pathParameterSubSchema"},
            ],
        },
        "parameter": {
            "oneOf": [
                {"$ref": "#/definitions/bodyParameter"},
                {"$ref": "#/definitions/nonBodyParameter"},
            ]
        },
        "schema": {
            "type": "object",
            "description": "A deterministic version of a JSON Schema object.",
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
            "properties": {
                "$ref": {"type": "string"},
                "format": {"type": "string"},
                "title": {
                    "$ref": "http://json-schema.org/draft-04/schema#/properties/title"
                },
                "description": {
                    "$ref": "http://json-schema.org/draft-04/schema#/properties/description"
                },
                "default": {
                    "$ref": "http://json-schema.org/draft-04/schema#/properties/default"
                },
                "multipleOf": {
                    "$ref": "http://json-schema.org/draft-04/schema#/properties/multipleOf"
                },
                "maximum": {
                    "$ref": "http://json-schema.org/draft-04/schema#/properties/maximum"
                },
                "exclusiveMaximum": {
                    "$ref": "http://json-schema.org/draft-04/schema#/properties/exclusiveMaximum"
                },
                "minimum": {
                    "$ref": "http://json-schema.org/draft-04/schema#/properties/minimum"
                },
                "exclusiveMinimum": {
                    "$ref": "http://json-schema.org/draft-04/schema#/properties/exclusiveMinimum"
                },
                "maxLength": {
                    "$ref": "http://json-schema.org/draft-04/schema#/definitions/positiveInteger"
                },
                "minLength": {
                    "$ref": "http://json-schema.org/draft-04/schema#/definitions/positiveIntegerDefault0"
                },
                "pattern": {
                    "$ref": "http://json-schema.org/draft-04/schema#/properties/pattern"
                },
                "maxItems": {
                    "$ref": "http://json-schema.org/draft-04/schema#/definitions/positiveInteger"
                },
                "minItems": {
                    "$ref": "http://json-schema.org/draft-04/schema#/definitions/positiveIntegerDefault0"
                },
                "uniqueItems": {
                    "$ref": "http://json-schema.org/draft-04/schema#/properties/uniqueItems"
                },
                "maxProperties": {
                    "$ref": "http://json-schema.org/draft-04/schema#/definitions/positiveInteger"
                },
                "minProperties": {
                    "$ref": "http://json-schema.org/draft-04/schema#/definitions/positiveIntegerDefault0"
                },
                "required": {
                    "$ref": "http://json-schema.org/draft-04/schema#/definitions/stringArray"
                },
                "enum": {
                    "$ref": "http://json-schema.org/draft-04/schema#/properties/enum"
                },
                "additionalProperties": {
                    "anyOf": [{"$ref": "#/definitions/schema"}, {"type": "boolean"}],
                    "default": {},
                },
                "type": {
                    "$ref": "http://json-schema.org/draft-04/schema#/properties/type"
                },
                "items": {
                    "anyOf": [
                        {"$ref": "#/definitions/schema"},
                        {
                            "type": "array",
                            "minItems": 1,
                            "items": {"$ref": "#/definitions/schema"},
                        },
                    ],
                    "default": {},
                },
                "allOf": {
                    "type": "array",
                    "minItems": 1,
                    "items": {"$ref": "#/definitions/schema"},
                },
                "properties": {
                    "type": "object",
                    "additionalProperties": {"$ref": "#/definitions/schema"},
                    "default": {},
                },
                "discriminator": {"type": "string"},
                "readOnly": {"type": "boolean", "default": False},
                "xml": {"$ref": "#/definitions/xml"},
                "externalDocs": {"$ref": "#/definitions/externalDocs"},
                "example": {},
            },
            "additionalProperties": False,
        },
        "fileSchema": {
            "type": "object",
            "description": "A deterministic version of a JSON Schema object.",
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
            "required": ["type"],
            "properties": {
                "format": {"type": "string"},
                "title": {
                    "$ref": "http://json-schema.org/draft-04/schema#/properties/title"
                },
                "description": {
                    "$ref": "http://json-schema.org/draft-04/schema#/properties/description"
                },
                "default": {
                    "$ref": "http://json-schema.org/draft-04/schema#/properties/default"
                },
                "required": {
                    "$ref": "http://json-schema.org/draft-04/schema#/definitions/stringArray"
                },
                "type": {"type": "string", "enum": ["file"]},
                "readOnly": {"type": "boolean", "default": False},
                "externalDocs": {"$ref": "#/definitions/externalDocs"},
                "example": {},
            },
            "additionalProperties": False,
        },
        "primitivesItems": {
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["string", "number", "integer", "boolean", "array"],
                },
                "format": {"type": "string"},
                "items": {"$ref": "#/definitions/primitivesItems"},
                "collectionFormat": {"$ref": "#/definitions/collectionFormat"},
                "default": {"$ref": "#/definitions/default"},
                "maximum": {"$ref": "#/definitions/maximum"},
                "exclusiveMaximum": {"$ref": "#/definitions/exclusiveMaximum"},
                "minimum": {"$ref": "#/definitions/minimum"},
                "exclusiveMinimum": {"$ref": "#/definitions/exclusiveMinimum"},
                "maxLength": {"$ref": "#/definitions/maxLength"},
                "minLength": {"$ref": "#/definitions/minLength"},
                "pattern": {"$ref": "#/definitions/pattern"},
                "maxItems": {"$ref": "#/definitions/maxItems"},
                "minItems": {"$ref": "#/definitions/minItems"},
                "uniqueItems": {"$ref": "#/definitions/uniqueItems"},
                "enum": {"$ref": "#/definitions/enum"},
                "multipleOf": {"$ref": "#/definitions/multipleOf"},
            },
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
        },
        "security": {
            "type": "array",
            "items": {"$ref": "#/definitions/securityRequirement"},
            "uniqueItems": True,
        },
        "securityRequirement": {
            "type": "object",
            "additionalProperties": {
                "type": "array",
                "items": {"type": "string"},
                "uniqueItems": True,
            },
        },
        "xml": {
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "name": {"type": "string"},
                "namespace": {"type": "string"},
                "prefix": {"type": "string"},
                "attribute": {"type": "boolean", "default": False},
                "wrapped": {"type": "boolean", "default": False},
            },
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
        },
        "tag": {
            "type": "object",
            "additionalProperties": False,
            "required": ["name"],
            "properties": {
                "name": {"type": "string"},
                "description": {"type": "string"},
                "externalDocs": {"$ref": "#/definitions/externalDocs"},
            },
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
        },
        "securityDefinitions": {
            "type": "object",
            "additionalProperties": {
                "oneOf": [
                    {"$ref": "#/definitions/basicAuthenticationSecurity"},
                    {"$ref": "#/definitions/apiKeySecurity"},
                    {"$ref": "#/definitions/oauth2ImplicitSecurity"},
                    {"$ref": "#/definitions/oauth2PasswordSecurity"},
                    {"$ref": "#/definitions/oauth2ApplicationSecurity"},
                    {"$ref": "#/definitions/oauth2AccessCodeSecurity"},
                ]
            },
        },
        "basicAuthenticationSecurity": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {"type": "string", "enum": ["basic"]},
                "description": {"type": "string"},
            },
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
        },
        "apiKeySecurity": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type", "name", "in"],
            "properties": {
                "type": {"type": "string", "enum": ["apiKey"]},
                "name": {"type": "string"},
                "in": {"type": "string", "enum": ["header", "query"]},
                "description": {"type": "string"},
            },
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
        },
        "oauth2ImplicitSecurity": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type", "flow", "authorizationUrl"],
            "properties": {
                "type": {"type": "string", "enum": ["oauth2"]},
                "flow": {"type": "string", "enum": ["implicit"]},
                "scopes": {"$ref": "#/definitions/oauth2Scopes"},
                "authorizationUrl": {"type": "string", "format": "uri"},
                "description": {"type": "string"},
            },
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
        },
        "oauth2PasswordSecurity": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type", "flow", "tokenUrl"],
            "properties": {
                "type": {"type": "string", "enum": ["oauth2"]},
                "flow": {"type": "string", "enum": ["password"]},
                "scopes": {"$ref": "#/definitions/oauth2Scopes"},
                "tokenUrl": {"type": "string", "format": "uri"},
                "description": {"type": "string"},
            },
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
        },
        "oauth2ApplicationSecurity": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type", "flow", "tokenUrl"],
            "properties": {
                "type": {"type": "string", "enum": ["oauth2"]},
                "flow": {"type": "string", "enum": ["application"]},
                "scopes": {"$ref": "#/definitions/oauth2Scopes"},
                "tokenUrl": {"type": "string", "format": "uri"},
                "description": {"type": "string"},
            },
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
        },
        "oauth2AccessCodeSecurity": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type", "flow", "authorizationUrl", "tokenUrl"],
            "properties": {
                "type": {"type": "string", "enum": ["oauth2"]},
                "flow": {"type": "string", "enum": ["accessCode"]},
                "scopes": {"$ref": "#/definitions/oauth2Scopes"},
                "authorizationUrl": {"type": "string", "format": "uri"},
                "tokenUrl": {"type": "string", "format": "uri"},
                "description": {"type": "string"},
            },
            "patternProperties": {"^x-": {"$ref": "#/definitions/vendorExtension"}},
        },
        "oauth2Scopes": {"type": "object", "additionalProperties": {"type": "string"}},
        "mediaTypeList": {
            "type": "array",
            "items": {"$ref": "#/definitions/mimeType"},
            "uniqueItems": True,
        },
        "parametersList": {
            "type": "array",
            "description": "The parameters needed to send a valid API call.",
            "additionalItems": False,
            "items": {
                "oneOf": [
                    {"$ref": "#/definitions/parameter"},
                    {"$ref": "#/definitions/jsonReference"},
                ]
            },
            "uniqueItems": True,
        },
        "schemesList": {
            "type": "array",
            "description": "The transfer protocol of the API.",
            "items": {"type": "string", "enum": ["http", "https", "ws", "wss"]},
            "uniqueItems": True,
        },
        "collectionFormat": {
            "type": "string",
            "enum": ["csv", "ssv", "tsv", "pipes"],
            "default": "csv",
        },
        "collectionFormatWithMulti": {
            "type": "string",
            "enum": ["csv", "ssv", "tsv", "pipes", "multi"],
            "default": "csv",
        },
        "title": {"$ref": "http://json-schema.org/draft-04/schema#/properties/title"},
        "description": {
            "$ref": "http://json-schema.org/draft-04/schema#/properties/description"
        },
        "default": {
            "$ref": "http://json-schema.org/draft-04/schema#/properties/default"
        },
        "multipleOf": {
            "$ref": "http://json-schema.org/draft-04/schema#/properties/multipleOf"
        },
        "maximum": {
            "$ref": "http://json-schema.org/draft-04/schema#/properties/maximum"
        },
        "exclusiveMaximum": {
            "$ref": "http://json-schema.org/draft-04/schema#/properties/exclusiveMaximum"
        },
        "minimum": {
            "$ref": "http://json-schema.org/draft-04/schema#/properties/minimum"
        },
        "exclusiveMinimum": {
            "$ref": "http://json-schema.org/draft-04/schema#/properties/exclusiveMinimum"
        },
        "maxLength": {
            "$ref": "http://json-schema.org/draft-04/schema#/definitions/positiveInteger"
        },
        "minLength": {
            "$ref": "http://json-schema.org/draft-04/schema#/definitions/positiveIntegerDefault0"
        },
        "pattern": {
            "$ref": "http://json-schema.org/draft-04/schema#/properties/pattern"
        },
        "maxItems": {
            "$ref": "http://json-schema.org/draft-04/schema#/definitions/positiveInteger"
        },
        "minItems": {
            "$ref": "http://json-schema.org/draft-04/schema#/definitions/positiveIntegerDefault0"
        },
        "uniqueItems": {
            "$ref": "http://json-schema.org/draft-04/schema#/properties/uniqueItems"
        },
        "enum": {"$ref": "http://json-schema.org/draft-04/schema#/properties/enum"},
        "jsonReference": {
            "type": "object",
            "required": ["$ref"],
            "additionalProperties": False,
            "properties": {"$ref": {"type": "string"}},
        },
    },
}

SWAGGER_V3_JSONSCHEMA = {
    "$id": "https://spec.openapis.org/oas/3.1/schema/2022-10-07",
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "description": "The description of OpenAPI v3.1.x documents without schema validation, as defined by https://spec.openapis.org/oas/v3.1.0",
    "type": "object",
    "properties": {
        "openapi": {"type": "string", "pattern": "^3\\.1\\.\\d+(-.+)?$"},
        "info": {"$ref": "#/$defs/info"},
        "jsonSchemaDialect": {
            "type": "string",
            "format": "uri",
            "default": "https://spec.openapis.org/oas/3.1/dialect/base",
        },
        "servers": {
            "type": "array",
            "items": {"$ref": "#/$defs/server"},
            "default": [{"url": "/"}],
        },
        "paths": {"$ref": "#/$defs/paths"},
        "webhooks": {
            "type": "object",
            "additionalProperties": {"$ref": "#/$defs/path-item-or-reference"},
        },
        "components": {"$ref": "#/$defs/components"},
        "security": {
            "type": "array",
            "items": {"$ref": "#/$defs/security-requirement"},
        },
        "tags": {"type": "array", "items": {"$ref": "#/$defs/tag"}},
        "externalDocs": {"$ref": "#/$defs/external-documentation"},
    },
    "required": ["openapi", "info"],
    "anyOf": [
        {"required": ["paths"]},
        {"required": ["components"]},
        {"required": ["webhooks"]},
    ],
    "$ref": "#/$defs/specification-extensions",
    "unevaluatedProperties": False,
    "$defs": {
        "info": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#info-object",
            "type": "object",
            "properties": {
                "title": {"type": "string"},
                "summary": {"type": "string"},
                "description": {"type": "string"},
                "termsOfService": {"type": "string", "format": "uri"},
                "contact": {"$ref": "#/$defs/contact"},
                "license": {"$ref": "#/$defs/license"},
                "version": {"type": "string"},
            },
            "required": ["title", "version"],
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
        },
        "contact": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#contact-object",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "url": {"type": "string", "format": "uri"},
                "email": {"type": "string", "format": "email"},
            },
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
        },
        "license": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#license-object",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "identifier": {"type": "string"},
                "url": {"type": "string", "format": "uri"},
            },
            "required": ["name"],
            "dependentSchemas": {"identifier": {"not": {"required": ["url"]}}},
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
        },
        "server": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#server-object",
            "type": "object",
            "properties": {
                "url": {"type": "string", "format": "uri-reference"},
                "description": {"type": "string"},
                "variables": {
                    "type": "object",
                    "additionalProperties": {"$ref": "#/$defs/server-variable"},
                },
            },
            "required": ["url"],
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
        },
        "server-variable": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#server-variable-object",
            "type": "object",
            "properties": {
                "enum": {"type": "array", "items": {"type": "string"}, "minItems": 1},
                "default": {"type": "string"},
                "description": {"type": "string"},
            },
            "required": ["default"],
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
        },
        "components": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#components-object",
            "type": "object",
            "properties": {
                "schemas": {
                    "type": "object",
                    "additionalProperties": {"$dynamicRef": "#meta"},
                },
                "responses": {
                    "type": "object",
                    "additionalProperties": {"$ref": "#/$defs/response-or-reference"},
                },
                "parameters": {
                    "type": "object",
                    "additionalProperties": {"$ref": "#/$defs/parameter-or-reference"},
                },
                "examples": {
                    "type": "object",
                    "additionalProperties": {"$ref": "#/$defs/example-or-reference"},
                },
                "requestBodies": {
                    "type": "object",
                    "additionalProperties": {
                        "$ref": "#/$defs/request-body-or-reference"
                    },
                },
                "headers": {
                    "type": "object",
                    "additionalProperties": {"$ref": "#/$defs/header-or-reference"},
                },
                "securitySchemes": {
                    "type": "object",
                    "additionalProperties": {
                        "$ref": "#/$defs/security-scheme-or-reference"
                    },
                },
                "links": {
                    "type": "object",
                    "additionalProperties": {"$ref": "#/$defs/link-or-reference"},
                },
                "callbacks": {
                    "type": "object",
                    "additionalProperties": {"$ref": "#/$defs/callbacks-or-reference"},
                },
                "pathItems": {
                    "type": "object",
                    "additionalProperties": {"$ref": "#/$defs/path-item-or-reference"},
                },
            },
            "patternProperties": {
                "^(schemas|responses|parameters|examples|requestBodies|headers|securitySchemes|links|callbacks|pathItems)$": {
                    "$comment": "Enumerating all of the property names in the regex above is necessary for unevaluatedProperties to work as expected",
                    "propertyNames": {"pattern": "^[a-zA-Z0-9._-]+$"},
                }
            },
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
        },
        "paths": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#paths-object",
            "type": "object",
            "patternProperties": {"^/": {"$ref": "#/$defs/path-item"}},
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
        },
        "path-item": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#path-item-object",
            "type": "object",
            "properties": {
                "summary": {"type": "string"},
                "description": {"type": "string"},
                "servers": {"type": "array", "items": {"$ref": "#/$defs/server"}},
                "parameters": {
                    "type": "array",
                    "items": {"$ref": "#/$defs/parameter-or-reference"},
                },
                "get": {"$ref": "#/$defs/operation"},
                "put": {"$ref": "#/$defs/operation"},
                "post": {"$ref": "#/$defs/operation"},
                "delete": {"$ref": "#/$defs/operation"},
                "options": {"$ref": "#/$defs/operation"},
                "head": {"$ref": "#/$defs/operation"},
                "patch": {"$ref": "#/$defs/operation"},
                "trace": {"$ref": "#/$defs/operation"},
            },
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
        },
        "path-item-or-reference": {
            "if": {"type": "object", "required": ["$ref"]},
            "then": {"$ref": "#/$defs/reference"},
            "else": {"$ref": "#/$defs/path-item"},
        },
        "operation": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#operation-object",
            "type": "object",
            "properties": {
                "tags": {"type": "array", "items": {"type": "string"}},
                "summary": {"type": "string"},
                "description": {"type": "string"},
                "externalDocs": {"$ref": "#/$defs/external-documentation"},
                "operationId": {"type": "string"},
                "parameters": {
                    "type": "array",
                    "items": {"$ref": "#/$defs/parameter-or-reference"},
                },
                "requestBody": {"$ref": "#/$defs/request-body-or-reference"},
                "responses": {"$ref": "#/$defs/responses"},
                "callbacks": {
                    "type": "object",
                    "additionalProperties": {"$ref": "#/$defs/callbacks-or-reference"},
                },
                "deprecated": {"default": False, "type": "boolean"},
                "security": {
                    "type": "array",
                    "items": {"$ref": "#/$defs/security-requirement"},
                },
                "servers": {"type": "array", "items": {"$ref": "#/$defs/server"}},
            },
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
        },
        "external-documentation": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#external-documentation-object",
            "type": "object",
            "properties": {
                "description": {"type": "string"},
                "url": {"type": "string", "format": "uri"},
            },
            "required": ["url"],
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
        },
        "parameter": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#parameter-object",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "in": {"enum": ["query", "header", "path", "cookie"]},
                "description": {"type": "string"},
                "required": {"default": False, "type": "boolean"},
                "deprecated": {"default": False, "type": "boolean"},
                "schema": {"$dynamicRef": "#meta"},
                "content": {
                    "$ref": "#/$defs/content",
                    "minProperties": 1,
                    "maxProperties": 1,
                },
            },
            "required": ["name", "in"],
            "oneOf": [{"required": ["schema"]}, {"required": ["content"]}],
            "if": {"properties": {"in": {"const": "query"}}, "required": ["in"]},
            "then": {
                "properties": {"allowEmptyValue": {"default": False, "type": "boolean"}}
            },
            "dependentSchemas": {
                "schema": {
                    "properties": {
                        "style": {"type": "string"},
                        "explode": {"type": "boolean"},
                    },
                    "allOf": [
                        {"$ref": "#/$defs/examples"},
                        {
                            "$ref": "#/$defs/parameter/dependentSchemas/schema/$defs/styles-for-path"
                        },
                        {
                            "$ref": "#/$defs/parameter/dependentSchemas/schema/$defs/styles-for-header"
                        },
                        {
                            "$ref": "#/$defs/parameter/dependentSchemas/schema/$defs/styles-for-query"
                        },
                        {
                            "$ref": "#/$defs/parameter/dependentSchemas/schema/$defs/styles-for-cookie"
                        },
                        {
                            "$ref": "#/$defs/parameter/dependentSchemas/schema/$defs/styles-for-form"
                        },
                    ],
                    "$defs": {
                        "styles-for-path": {
                            "if": {
                                "properties": {"in": {"const": "path"}},
                                "required": ["in"],
                            },
                            "then": {
                                "properties": {
                                    "name": {"pattern": "[^/#?]+$"},
                                    "style": {
                                        "default": "simple",
                                        "enum": ["matrix", "label", "simple"],
                                    },
                                    "required": {"const": True},
                                },
                                "required": ["required"],
                            },
                        },
                        "styles-for-header": {
                            "if": {
                                "properties": {"in": {"const": "header"}},
                                "required": ["in"],
                            },
                            "then": {
                                "properties": {
                                    "style": {"default": "simple", "const": "simple"}
                                }
                            },
                        },
                        "styles-for-query": {
                            "if": {
                                "properties": {"in": {"const": "query"}},
                                "required": ["in"],
                            },
                            "then": {
                                "properties": {
                                    "style": {
                                        "default": "form",
                                        "enum": [
                                            "form",
                                            "spaceDelimited",
                                            "pipeDelimited",
                                            "deepObject",
                                        ],
                                    },
                                    "allowReserved": {
                                        "default": False,
                                        "type": "boolean",
                                    },
                                }
                            },
                        },
                        "styles-for-cookie": {
                            "if": {
                                "properties": {"in": {"const": "cookie"}},
                                "required": ["in"],
                            },
                            "then": {
                                "properties": {
                                    "style": {"default": "form", "const": "form"}
                                }
                            },
                        },
                        "styles-for-form": {
                            "if": {
                                "properties": {"style": {"const": "form"}},
                                "required": ["style"],
                            },
                            "then": {"properties": {"explode": {"default": True}}},
                            "else": {"properties": {"explode": {"default": False}}},
                        },
                    },
                }
            },
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
        },
        "parameter-or-reference": {
            "if": {"type": "object", "required": ["$ref"]},
            "then": {"$ref": "#/$defs/reference"},
            "else": {"$ref": "#/$defs/parameter"},
        },
        "request-body": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#request-body-object",
            "type": "object",
            "properties": {
                "description": {"type": "string"},
                "content": {"$ref": "#/$defs/content"},
                "required": {"default": False, "type": "boolean"},
            },
            "required": ["content"],
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
        },
        "request-body-or-reference": {
            "if": {"type": "object", "required": ["$ref"]},
            "then": {"$ref": "#/$defs/reference"},
            "else": {"$ref": "#/$defs/request-body"},
        },
        "content": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#fixed-fields-10",
            "type": "object",
            "additionalProperties": {"$ref": "#/$defs/media-type"},
            "propertyNames": {"format": "media-range"},
        },
        "media-type": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#media-type-object",
            "type": "object",
            "properties": {
                "schema": {"$dynamicRef": "#meta"},
                "encoding": {
                    "type": "object",
                    "additionalProperties": {"$ref": "#/$defs/encoding"},
                },
            },
            "allOf": [
                {"$ref": "#/$defs/specification-extensions"},
                {"$ref": "#/$defs/examples"},
            ],
            "unevaluatedProperties": False,
        },
        "encoding": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#encoding-object",
            "type": "object",
            "properties": {
                "contentType": {"type": "string", "format": "media-range"},
                "headers": {
                    "type": "object",
                    "additionalProperties": {"$ref": "#/$defs/header-or-reference"},
                },
                "style": {
                    "default": "form",
                    "enum": ["form", "spaceDelimited", "pipeDelimited", "deepObject"],
                },
                "explode": {"type": "boolean"},
                "allowReserved": {"default": False, "type": "boolean"},
            },
            "allOf": [
                {"$ref": "#/$defs/specification-extensions"},
                {"$ref": "#/$defs/encoding/$defs/explode-default"},
            ],
            "unevaluatedProperties": False,
            "$defs": {
                "explode-default": {
                    "if": {
                        "properties": {"style": {"const": "form"}},
                        "required": ["style"],
                    },
                    "then": {"properties": {"explode": {"default": True}}},
                    "else": {"properties": {"explode": {"default": False}}},
                }
            },
        },
        "responses": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#responses-object",
            "type": "object",
            "properties": {"default": {"$ref": "#/$defs/response-or-reference"}},
            "patternProperties": {
                "^[1-5](?:[0-9]{2}|XX)$": {"$ref": "#/$defs/response-or-reference"}
            },
            "minProperties": 1,
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
            "if": {
                "$comment": "either default, or at least one response code property must exist",
                "patternProperties": {"^[1-5](?:[0-9]{2}|XX)$": False},
            },
            "then": {"required": ["default"]},
        },
        "response": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#response-object",
            "type": "object",
            "properties": {
                "description": {"type": "string"},
                "headers": {
                    "type": "object",
                    "additionalProperties": {"$ref": "#/$defs/header-or-reference"},
                },
                "content": {"$ref": "#/$defs/content"},
                "links": {
                    "type": "object",
                    "additionalProperties": {"$ref": "#/$defs/link-or-reference"},
                },
            },
            "required": ["description"],
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
        },
        "response-or-reference": {
            "if": {"type": "object", "required": ["$ref"]},
            "then": {"$ref": "#/$defs/reference"},
            "else": {"$ref": "#/$defs/response"},
        },
        "callbacks": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#callback-object",
            "type": "object",
            "$ref": "#/$defs/specification-extensions",
            "additionalProperties": {"$ref": "#/$defs/path-item-or-reference"},
        },
        "callbacks-or-reference": {
            "if": {"type": "object", "required": ["$ref"]},
            "then": {"$ref": "#/$defs/reference"},
            "else": {"$ref": "#/$defs/callbacks"},
        },
        "example": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#example-object",
            "type": "object",
            "properties": {
                "summary": {"type": "string"},
                "description": {"type": "string"},
                "value": True,
                "externalValue": {"type": "string", "format": "uri"},
            },
            "not": {"required": ["value", "externalValue"]},
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
        },
        "example-or-reference": {
            "if": {"type": "object", "required": ["$ref"]},
            "then": {"$ref": "#/$defs/reference"},
            "else": {"$ref": "#/$defs/example"},
        },
        "link": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#link-object",
            "type": "object",
            "properties": {
                "operationRef": {"type": "string", "format": "uri-reference"},
                "operationId": {"type": "string"},
                "parameters": {"$ref": "#/$defs/map-of-strings"},
                "requestBody": True,
                "description": {"type": "string"},
                "body": {"$ref": "#/$defs/server"},
            },
            "oneOf": [{"required": ["operationRef"]}, {"required": ["operationId"]}],
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
        },
        "link-or-reference": {
            "if": {"type": "object", "required": ["$ref"]},
            "then": {"$ref": "#/$defs/reference"},
            "else": {"$ref": "#/$defs/link"},
        },
        "header": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#header-object",
            "type": "object",
            "properties": {
                "description": {"type": "string"},
                "required": {"default": False, "type": "boolean"},
                "deprecated": {"default": False, "type": "boolean"},
                "schema": {"$dynamicRef": "#meta"},
                "content": {
                    "$ref": "#/$defs/content",
                    "minProperties": 1,
                    "maxProperties": 1,
                },
            },
            "oneOf": [{"required": ["schema"]}, {"required": ["content"]}],
            "dependentSchemas": {
                "schema": {
                    "properties": {
                        "style": {"default": "simple", "const": "simple"},
                        "explode": {"default": False, "type": "boolean"},
                    },
                    "$ref": "#/$defs/examples",
                }
            },
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
        },
        "header-or-reference": {
            "if": {"type": "object", "required": ["$ref"]},
            "then": {"$ref": "#/$defs/reference"},
            "else": {"$ref": "#/$defs/header"},
        },
        "tag": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#tag-object",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "description": {"type": "string"},
                "externalDocs": {"$ref": "#/$defs/external-documentation"},
            },
            "required": ["name"],
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
        },
        "reference": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#reference-object",
            "type": "object",
            "properties": {
                "$ref": {"type": "string", "format": "uri-reference"},
                "summary": {"type": "string"},
                "description": {"type": "string"},
            },
            "unevaluatedProperties": False,
        },
        "schema": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#schema-object",
            "$dynamicAnchor": "meta",
            "type": ["object", "boolean"],
        },
        "security-scheme": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#security-scheme-object",
            "type": "object",
            "properties": {
                "type": {
                    "enum": ["apiKey", "http", "mutualTLS", "oauth2", "openIdConnect"]
                },
                "description": {"type": "string"},
            },
            "required": ["type"],
            "allOf": [
                {"$ref": "#/$defs/specification-extensions"},
                {"$ref": "#/$defs/security-scheme/$defs/type-apikey"},
                {"$ref": "#/$defs/security-scheme/$defs/type-http"},
                {"$ref": "#/$defs/security-scheme/$defs/type-http-bearer"},
                {"$ref": "#/$defs/security-scheme/$defs/type-oauth2"},
                {"$ref": "#/$defs/security-scheme/$defs/type-oidc"},
            ],
            "unevaluatedProperties": False,
            "$defs": {
                "type-apikey": {
                    "if": {
                        "properties": {"type": {"const": "apiKey"}},
                        "required": ["type"],
                    },
                    "then": {
                        "properties": {
                            "name": {"type": "string"},
                            "in": {"enum": ["query", "header", "cookie"]},
                        },
                        "required": ["name", "in"],
                    },
                },
                "type-http": {
                    "if": {
                        "properties": {"type": {"const": "http"}},
                        "required": ["type"],
                    },
                    "then": {
                        "properties": {"scheme": {"type": "string"}},
                        "required": ["scheme"],
                    },
                },
                "type-http-bearer": {
                    "if": {
                        "properties": {
                            "type": {"const": "http"},
                            "scheme": {
                                "type": "string",
                                "pattern": "^[Bb][Ee][Aa][Rr][Ee][Rr]$",
                            },
                        },
                        "required": ["type", "scheme"],
                    },
                    "then": {"properties": {"bearerFormat": {"type": "string"}}},
                },
                "type-oauth2": {
                    "if": {
                        "properties": {"type": {"const": "oauth2"}},
                        "required": ["type"],
                    },
                    "then": {
                        "properties": {"flows": {"$ref": "#/$defs/oauth-flows"}},
                        "required": ["flows"],
                    },
                },
                "type-oidc": {
                    "if": {
                        "properties": {"type": {"const": "openIdConnect"}},
                        "required": ["type"],
                    },
                    "then": {
                        "properties": {
                            "openIdConnectUrl": {"type": "string", "format": "uri"}
                        },
                        "required": ["openIdConnectUrl"],
                    },
                },
            },
        },
        "security-scheme-or-reference": {
            "if": {"type": "object", "required": ["$ref"]},
            "then": {"$ref": "#/$defs/reference"},
            "else": {"$ref": "#/$defs/security-scheme"},
        },
        "oauth-flows": {
            "type": "object",
            "properties": {
                "implicit": {"$ref": "#/$defs/oauth-flows/$defs/implicit"},
                "password": {"$ref": "#/$defs/oauth-flows/$defs/password"},
                "clientCredentials": {
                    "$ref": "#/$defs/oauth-flows/$defs/client-credentials"
                },
                "authorizationCode": {
                    "$ref": "#/$defs/oauth-flows/$defs/authorization-code"
                },
            },
            "$ref": "#/$defs/specification-extensions",
            "unevaluatedProperties": False,
            "$defs": {
                "implicit": {
                    "type": "object",
                    "properties": {
                        "authorizationUrl": {"type": "string", "format": "uri"},
                        "refreshUrl": {"type": "string", "format": "uri"},
                        "scopes": {"$ref": "#/$defs/map-of-strings"},
                    },
                    "required": ["authorizationUrl", "scopes"],
                    "$ref": "#/$defs/specification-extensions",
                    "unevaluatedProperties": False,
                },
                "password": {
                    "type": "object",
                    "properties": {
                        "tokenUrl": {"type": "string", "format": "uri"},
                        "refreshUrl": {"type": "string", "format": "uri"},
                        "scopes": {"$ref": "#/$defs/map-of-strings"},
                    },
                    "required": ["tokenUrl", "scopes"],
                    "$ref": "#/$defs/specification-extensions",
                    "unevaluatedProperties": False,
                },
                "client-credentials": {
                    "type": "object",
                    "properties": {
                        "tokenUrl": {"type": "string", "format": "uri"},
                        "refreshUrl": {"type": "string", "format": "uri"},
                        "scopes": {"$ref": "#/$defs/map-of-strings"},
                    },
                    "required": ["tokenUrl", "scopes"],
                    "$ref": "#/$defs/specification-extensions",
                    "unevaluatedProperties": False,
                },
                "authorization-code": {
                    "type": "object",
                    "properties": {
                        "authorizationUrl": {"type": "string", "format": "uri"},
                        "tokenUrl": {"type": "string", "format": "uri"},
                        "refreshUrl": {"type": "string", "format": "uri"},
                        "scopes": {"$ref": "#/$defs/map-of-strings"},
                    },
                    "required": ["authorizationUrl", "tokenUrl", "scopes"],
                    "$ref": "#/$defs/specification-extensions",
                    "unevaluatedProperties": False,
                },
            },
        },
        "security-requirement": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#security-requirement-object",
            "type": "object",
            "additionalProperties": {"type": "array", "items": {"type": "string"}},
        },
        "specification-extensions": {
            "$comment": "https://spec.openapis.org/oas/v3.1.0#specification-extensions",
            "patternProperties": {"^x-": True},
        },
        "examples": {
            "properties": {
                "example": True,
                "examples": {
                    "type": "object",
                    "additionalProperties": {"$ref": "#/$defs/example-or-reference"},
                },
            }
        },
        "map-of-strings": {
            "type": "object",
            "additionalProperties": {"type": "string"},
        },
    },
}



import jsonschema
from typing import Any, Dict

from flask_rebar.testing.swagger_jsonschema import SWAGGER_V2_JSONSCHEMA


def validate_swagger(
    swagger: Dict[str, Any], schema: Dict[str, Any] = SWAGGER_V2_JSONSCHEMA
) -> None:
    """
    Validates that a dictionary is a valid Swagger spec.

    :param dict swagger: The swagger spec
    :param dict schema: The JSON Schema to use to validate the swagger spec
    :raises: jsonschema.ValidationError
    """
    jsonschema.validate(instance=swagger, schema=schema)



"""
    Swagger UI Blueprint
    ~~~~~~~~~~~~~~~~~~~~

    Flask Blueprint for adding Swagger UI to an API.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
from flask import Blueprint, render_template


def create_swagger_ui_blueprint(
    ui_url: str,
    swagger_url: str,
    name: str = "swagger_ui",
    page_title: str = "Swagger UI",
) -> Blueprint:
    """
    Create a blueprint for adding Swagger UI to a service.

    :param str ui_url:
        The path where the Swagger UI will be served from.
        All static files will be served from here as well.
    :param str swagger_url:
        The path (or full URL) where the Swagger UI can retrieve
        the Swagger specification.
    :param str name:
        A name for the blueprint. This is useful if the API is
        hosting multiple instances of the Swagger UI.
    :param str page_title:
        Name to use as the title for the HTML page.
    :rtype: flask.Blueprint
    """
    blueprint = Blueprint(
        name=name,
        import_name=__name__,
        static_folder="static",
        template_folder="templates",
        url_prefix=ui_url,
    )

    template_context = {
        "blueprint_name": name,
        "swagger_url": swagger_url,
        "page_title": page_title,
    }

    @blueprint.route("/")
    @blueprint.route("")
    def show() -> str:
        return render_template("index.html.jinja2", **template_context)

    return blueprint



"""
    Generator Utilities
    ~~~~~~~~~~~~~~~~~~~

    Helper functions shared by multiple Swagger Generators.

    :copyright: Copyright 2019 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
import copy
import re
from collections import namedtuple, OrderedDict
from typing import (
    overload,
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Set,
    Union,
    TYPE_CHECKING,
)

from flask_rebar.authenticators import Authenticator
from flask_rebar.utils.defaults import USE_DEFAULT
from flask_rebar.swagger_generation import swagger_words as sw
from flask_rebar.swagger_generation.marshmallow_to_swagger import get_swagger_title
from marshmallow import Schema

if TYPE_CHECKING:
    from flask_rebar.rebar import HandlerRegistry
    from flask_rebar.rebar import PathDefinition


def get_key(obj: Dict[str, Any]) -> str:
    """
    Returns the key for a JSONSchema object that we can use to make a $ref.

    We're just enforcing that objects all have a title for now.

    :param dict obj:
    :rtype: str
    """
    return obj[sw.title]


def create_ref(*parts: str) -> str:
    """Create a reference from `parts`

    For example:

        assert create_ref("#", "definitions", "foo") == "#/definitions/foo"

    :param Tuple[str] parts:
    :rtype: str
    """
    return "/".join(parts)


def get_ref_schema(base: str, schema: Schema) -> Dict[str, Any]:
    """Create a JSONSchema object that is a reference to `schema`

    :param base: base string for references, e.g. "#/definitions"
    :param schema:
    :return:
    """
    ref: Dict[str, Any] = {sw.ref: create_ref(base, get_swagger_title(schema))}
    return ref if not schema.many else {sw.type_: sw.array, sw.items: ref}


def get_response_description(schema: Schema) -> str:
    """
    Retrieves a response description from a Marshmallow schema.

    This is a required field, so we _have_ to give something back.

    :param marshmallow.Schema schema:
    :rtype: str
    """
    if schema.__doc__:
        return schema.__doc__
    else:
        return get_swagger_title(schema)


def flatten(schema: Dict[str, Any], base: str) -> Tuple[Dict[str, str], Dict[str, Any]]:
    """
    Recursively flattens a JSONSchema to a dictionary of keyed JSONSchemas,
    replacing nested objects with a reference to that object.


    Example input::
        {
          'type': 'object',
          'title': 'x',
          'properties': {
            'a': {
              'type': 'object',
              'title': 'y',
              'properties': {'b': {'type': 'integer'}}
            }
          }
        }

    Example output::
        {
          'x': {
            'type': 'object',
            'title': 'x',
            'properties': {
              'a': {'$ref': '#/definitions/y'}
            }
          },
          'y': {
            'type': 'object',
            'title': 'y',
            'properties': {'b': {'type': 'integer'}}
          }
        }

    This is useful for decomposing complex object generated from Marshmallow
    into a definitions object in Swagger.

    :param dict schema:
    :param str base: Base string for references, e.g. "#/definitions"
    :rtype: tuple(dict, dict)
    :returns: A tuple where the first item is the input object with any nested
    objects replaces with references, and the second item is the flattened
    definitions dictionary.
    """
    schema = copy.deepcopy(schema)
    definitions: Dict[str, Any] = {}
    schema = _flatten(schema=schema, definitions=definitions, base=base)
    return schema, definitions


def _flatten(
    schema: Dict[str, Any], definitions: Dict[str, Any], base: str
) -> Dict[str, str]:
    schema_type = schema.get(sw.type_)
    subschema_keyword = _get_subschema_keyword(schema)

    if schema_type == sw.object_:
        properties = schema.get(sw.properties, {})
        for key, prop in properties.items():
            properties[key] = _flatten(schema=prop, definitions=definitions, base=base)

    elif schema_type == sw.array:
        schema[sw.items] = _flatten(
            schema=schema[sw.items], definitions=definitions, base=base
        )

    elif subschema_keyword:
        subschemas = schema[subschema_keyword]
        for i, subschema in enumerate(subschemas):
            subschemas[i] = _flatten(
                schema=subschema, definitions=definitions, base=base
            )

    if sw.title in schema:
        definitions_key = get_key(schema)
        definitions[definitions_key] = schema
        schema = {sw.ref: create_ref(base, definitions_key)}

    return schema


def _get_subschema_keyword(schema: Dict[str, Any]) -> Optional[str]:
    for keyword in (sw.any_of, sw.one_of, sw.all_of):
        if keyword in schema:
            return keyword
    return None


_PATH_REGEX = re.compile("<((?P<type>.+?):)?(?P<name>.+?)>")
PathArgument = namedtuple("PathArgument", ["name", "type"])


def format_path_for_swagger(path: str) -> Tuple[str, Tuple[PathArgument, ...]]:
    """
    Flask and Swagger represent paths differently - this parses a Flask path
    to its Swagger form. This also extracts what the arguments in the flask
    path are, so we can represent them as parameters in Swagger.

    :param str path:
    :rtype: tuple(str, tuple(_PathArgument))
    """
    matches = list(_PATH_REGEX.finditer(path))

    args = tuple(
        PathArgument(name=match.group("name"), type=match.group("type") or "string")
        for match in matches
    )

    subbed_path = _PATH_REGEX.sub(
        repl=lambda match: "{{{}}}".format(match.group("name")), string=path
    )
    return subbed_path, args


def verify_parameters_are_the_same(
    a: List[Dict[str, Any]], b: List[Dict[str, Any]]
) -> None:
    def get_sort_key(parameter: Dict[str, Any]) -> str:
        return parameter[sw.name]

    sorted_a = sorted(a, key=get_sort_key)
    sorted_b = sorted(b, key=get_sort_key)

    if sorted_a != sorted_b:
        msg = (
            "Swagger generation does not support Flask url "
            "converters that map to different Swagger types!"
        )
        raise ValueError(msg)


def get_unique_schema_definitions(
    registry: "HandlerRegistry",
    base: str,
    default_response_schema: Schema,
    response_converter: Callable[[Schema], Dict[str, Any]],
    request_body_converter: Callable[[Schema], Dict[str, Any]],
) -> Dict[str, Any]:
    """Extract a map of unique schema definitions for all marshal schemas and responses schemas in `registry`

    :param flask_rebar.rebar.HandlerRegistry registry:
    :param str base: Base string for references, e.g. "#/definitions"
    :param marshmallow.Schema default_response_schema:
    :param typing.Callable[[marshmallow.Schema], dict] response_converter:
    :param typing.Callable[[marshmallow.Schema], dict] request_body_converter:
    :rtype: dict
    """
    all_schemas = set()

    converted = []

    all_schemas.add(default_response_schema)
    converted.append(response_converter(default_response_schema))

    for d in iterate_path_definitions(paths=registry.paths):
        if d.response_body_schema:
            for schema in d.response_body_schema.values():
                if schema is None:
                    # Responses that don't have a response body have None
                    # for a schema
                    continue

                if schema not in all_schemas:
                    converted.append(response_converter(schema))
                all_schemas.add(schema)

        if d.request_body_schema:
            schema = d.request_body_schema

            if schema not in all_schemas:
                converted.append(request_body_converter(schema))

            all_schemas.add(schema)

    flattened = {}

    for obj in converted:
        _, flattened_definitions = flatten(obj, base)
        flattened.update(flattened_definitions)

    return flattened


def get_unique_authenticators(registry: "HandlerRegistry") -> Set[Authenticator]:
    authenticators = {
        authenticator
        for d in iterate_path_definitions(paths=registry.paths)
        for authenticator in d.authenticators
        if authenticator is not None and authenticator is not USE_DEFAULT
    }

    for authenticator in registry.default_authenticators:
        if authenticator is not None:
            authenticators.add(authenticator)

    return authenticators


def iterate_path_definitions(
    paths: Dict[str, Dict[str, "PathDefinition"]]
) -> Iterator["PathDefinition"]:
    """Iterate over all `PathDefinition` instances in `paths`

    :param dict[str, dict[str, flask_rebar.rebar.PathDefinition]] paths:
    :return Iterator[PathDefinition]
    """
    for methods in paths.values():
        yield from methods.values()


@overload
def recursively_convert_dict_to_ordered_dict(obj: Dict) -> OrderedDict:
    ...


@overload
def recursively_convert_dict_to_ordered_dict(obj: List[Dict]) -> List[OrderedDict]:
    ...


@overload
def recursively_convert_dict_to_ordered_dict(
    obj: List[OrderedDict],
) -> List[OrderedDict]:
    ...


def recursively_convert_dict_to_ordered_dict(
    obj: Union[Dict, List[Dict], List[OrderedDict]]
) -> Union[OrderedDict, List[OrderedDict]]:
    """Recursively converts a dictionary into and OrderedDict"""
    if isinstance(obj, dict):
        sorted_dict = OrderedDict(sorted(obj.items(), key=lambda i: i[0]))
        for key, val in obj.items():
            sorted_dict[key] = recursively_convert_dict_to_ordered_dict(val)
        return sorted_dict
    elif isinstance(obj, list):
        return [recursively_convert_dict_to_ordered_dict(item) for item in obj]
    else:
        return obj



from typing import Any
from typing import Dict

import marshmallow
from marshmallow.fields import Field
from marshmallow.schema import Schema

from flask import current_app
from flask_rebar.validation import filter_dump_only, RequireOnDumpMixin


def set_data_key(field: Field, key: str) -> Field:
    field.data_key = key
    return field


def get_data_key(field: Field) -> str:
    if field.data_key:
        return field.data_key
    if field.name is None:
        raise ValueError("Field name cannot be None")
    return field.name


def load(schema: Schema, data: Dict[str, Any]) -> Dict[str, Any]:
    return schema.load(data)


def dump(schema: Schema, data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Our wrapper for Schema.dump that includes optional validation.
    Note that as of Flask-Rebar 2.x (hence Marshmallow 3.x), Marshmallow's default behavior is to NOT validate on dump
    Accordingly, we are making validation "opt-in" here, which can be controlled at schema level with
    RequireOnDumpMixin or globally via validate_on_dump attribute of Rebar instance
    """
    try:
        force_validation = current_app.extensions["rebar"]["instance"].validate_on_dump
    except (
        RuntimeError
    ):  # running outside app context (some unit test cases, potentially ad hoc scripts)
        force_validation = False

    if isinstance(schema, RequireOnDumpMixin) or force_validation:
        try:
            # We do an initial schema.dump here in order to support arbitrary data objects (e.g., ORM objects, etc.)
            # and give us something we can pass to .load below
            # Since marshmallow 3 doesn't validate on dump, this has the effect of stripping unknown fields.
            result = schema.dump(data)
        except marshmallow.ValidationError:
            raise
        except Exception as e:
            raise marshmallow.ValidationError(str(e))

        # filter out "dump_only" fields before we call load - we are only calling load to validate data we are dumping
        # (We use load because that's how Marshmallow docs recommend doing this sort of validation, presumably because
        # @pre_load massaging of data could make otherwise invalid data valid.
        filtered = filter_dump_only(schema, result)
        schema.load(filtered.loadable)  # trigger validation
    else:
        result = schema.dump(data)
    return result


def exclude_unknown_fields(schema: Schema) -> Schema:
    schema.unknown = marshmallow.EXCLUDE
    return schema



from typing import Any, Optional, Type
from marshmallow import Schema

try:
    import marshmallow_objects as mo

    MARSHMALLOW_OBJECTS = True
except ImportError:
    MARSHMALLOW_OBJECTS = False


def get_marshmallow_objects_schema(model: Any) -> Optional[Type[Schema]]:
    if MARSHMALLOW_OBJECTS and (
        isinstance(model, mo.Model) or issubclass(model, mo.Model)
    ):
        return model.__get_schema_class__()
    else:
        return None


if MARSHMALLOW_OBJECTS:

    class NestedTitledModel(mo.NestedModel):
        """
        Use this class instead of marshmallow_object.NestedModel if you need to supply
        __swagger_title__ to override the default of {MyModelClass}Schema
        """

        def __init__(self, nested: Type[mo.Model], title: str, **kwargs: Any) -> None:
            super().__init__(nested, **kwargs)
            self.schema.__swagger_title__ = title

else:

    class NestedTitledModel:  # type: ignore
        """
        This version of NestedTitledModel will exist if marshmallow-objects is not present
        """

        def __init__(self) -> None:
            raise ImportError(
                "To use NestedTitledModel you must install marshmallow-objects"
            )



from flask_rebar.utils.request_utils import marshal, response

from flask_rebar.rebar import (
    Rebar,
    HandlerRegistry,
    get_validated_args,
    get_validated_body,
    get_validated_headers,
)

from flask_rebar.authenticators import HeaderApiKeyAuthenticator

from flask_rebar.validation import ResponseSchema, RequestSchema

from flask_rebar.swagger_generation.swagger_generator_v2 import SwaggerV2Generator
from flask_rebar.swagger_generation.swagger_generator_v3 import SwaggerV3Generator
from flask_rebar.swagger_generation.swagger_objects import (
    ExternalDocumentation,
    Tag,
    Server,
    ServerVariable,
)



#!/usr/bin/env python
from setuptools import setup, find_packages


# packages required for local development and testing
development = [
    "black==23.7.0",
    "bumpversion==0.6.0",
    "click>=8.1.3,<9.0.0",
    "flake8==6.0.0",
    "gitchangelog>=3.0.4,<4.0.0",
    "jsonschema==4.18.4",
    "marshmallow-objects~=2.3",
    "mypy==1.8.0",
    "parametrize==0.1.1",
    "pre-commit>=1.14.4",
    "pytest~=7.4",
    "pytest-order~=1.0",
    "Sphinx>=6.0.0,<7.0.0",
    "sphinx_rtd_theme==1.2.2",
    "types-jsonschema==4.17.0.10",
    "types-setuptools==68.0.0.3",
]

install_requires = [
    "Flask>=1.0,<4",
    "marshmallow>=3.0,<4",
    "typing-extensions>=4.8,<5;python_version<'3.10'",
    "Werkzeug>=2.2,<4",
]

if __name__ == "__main__":
    setup(
        name="flask-rebar",
        version="3.2.1",
        author="Barak Alon",
        author_email="barak.s.alon@gmail.com",
        description="Flask-Rebar combines flask, marshmallow, and swagger for robust REST services.",
        long_description=open("README.rst").read(),
        keywords=["flask", "rest", "marshmallow", "openapi", "swagger"],
        license="MIT",
        packages=find_packages(exclude=("test*", "examples")),
        package_data={"flask_rebar": ["py.typed"]},
        include_package_data=True,
        extras_require={"dev": development, "enum": ["marshmallow-enum~=1.5"]},
        install_requires=install_requires,
        url="https://github.com/plangrid/flask-rebar",
        classifiers=[
            "Environment :: Web Environment",
            "Framework :: Flask",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
            "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
            "Topic :: Software Development :: Libraries :: Application Frameworks",
            "Topic :: Software Development :: Libraries :: Python Modules",
        ],
    )



"""
    Base Authenticator
    ~~~~~~~~~~~~~~~~~~

    Base class for authenticators.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""


class Authenticator:
    """
    Abstract authenticator class. Custom authentication methods should
    extend this class.
    """

    def authenticate(self) -> None:
        """
        Implementations of :class:`Authenticator` should override this method.

        This will be called before a request handler is called, and should raise
        an :class:`flask_rebar.errors.HttpJsonError` is authentication fails.

        Otherwise the return value is ignored.

        :raises: :class:`flask_rebar.errors.Unauthorized`
        """
        raise NotImplementedError



"""
    Validation
    ~~~~~~~~~~

    Helpful extensions for Marshmallow objects.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
from collections import namedtuple
from typing import Any, List, Optional, Mapping, Union

from marshmallow import Schema
from marshmallow import fields
from werkzeug.datastructures import MultiDict

FilterResult = namedtuple("FilterResult", "loadable, dump_only")


def filter_dump_only(
    schema: Schema, data: Union[Mapping[str, Any], List[Mapping[str, Any]]]
) -> FilterResult:
    """
    Return a filtered copy of data in which any items matching a "dump_only" field are separated
    :param schema: Instance of a Schema class
    :param data: Mapping or List of Mappings with data
    :return: FilterResult
    """
    # Note as of marshmallow 3.13.0, Schema.dump_only is NOT populated if fields are declared as dump_only inline,
    # so we'll calculate "dump_only" ourselves.  ref: https://github.com/marshmallow-code/marshmallow/issues/1857
    output_to_input_keys = {
        field.data_key: key for key, field in schema.fields.items() if field.data_key
    }
    dump_only_fields = schema.dump_fields.keys() - schema.load_fields.keys()
    if isinstance(data, Mapping):
        dump_only = dict()
        non_dump_only = dict()
        # get our dump_only fields directly, and candidates for loadable:
        for k, v in data.items():
            if output_to_input_keys.get(k, k) in dump_only_fields:
                dump_only[k] = v
            else:
                non_dump_only[k] = v

        # construct loadable (a subset of non_dump_only, with recursive filter of nested dump_only fields)
        loadable = dict()
        for k, v in non_dump_only.items():
            field = schema.fields[output_to_input_keys.get(k, k)]
            # see if we have a nested schema (using either Nested(many=True) or List(Nested())
            field_schema = None
            if isinstance(field, fields.Nested):
                field_schema = field.schema
            elif isinstance(field, fields.List) and isinstance(
                field.inner, fields.Nested
            ):
                field_schema = field.inner.schema
            if field_schema is None:
                loadable[k] = v
            else:
                field_filtered = filter_dump_only(field_schema, v)
                loadable[k] = field_filtered.loadable
                dump_only[k] = field_filtered.dump_only
        return FilterResult(loadable=loadable, dump_only=dump_only)
    elif isinstance(data, list):
        processed_items = [filter_dump_only(schema, item) for item in data]
        return FilterResult(
            [item.loadable for item in processed_items],
            [item.dump_only for item in processed_items],
        )
    elif data is None:
        return FilterResult(loadable=dict(), dump_only=dict())
    else:
        # I am not aware of any case where we should get something other than a Mapping or list, but just in case
        # we can raise a hopefully helpful error if there's some weird Schema that can cause that, so we know
        # we need to update this and patch rebar ;)
        raise TypeError(f"filter_dump_only doesn't understand data type {type(data)}")


class CommaSeparatedList(fields.List):
    """
    A field class for Marshmallow; use this class when your list will be
    deserialized from a comma separated list of values.
    e.g. ?foo=bar,baz -> {'foo': ['bar', 'baz']}
    """

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs: Any,
    ) -> List[Any]:
        if not isinstance(value, list):
            value = value.split(",")
        return super()._deserialize(value, attr, data)

    def _serialize(
        self,
        value: List[Any],
        attr: Optional[str],
        obj: Mapping[str, Any],
        **kwargs: Any,
    ) -> Any:
        # this function clearly returns a str but the superclass fields.List returns the type
        # list[Any] | None which is not compatible. should we subclass a generic Field instead?
        items = super()._serialize(value, attr, obj)
        if items is None:
            return None
        return ",".join([str(i) for i in items])


class QueryParamList(fields.List):
    """
    A field class for Marshmallow; use this class when your list will be
    deserialized from a query string containing the same param multiple
    times where each param is an item in the list.
    e.g. ?foo=bar&foo=baz -> {'foo': ['bar', 'baz']}
    """

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs: Any,
    ) -> List[Any]:
        # data is a MultiDict of query params, so pull out all of the items
        # with getlist instead of just the first
        if not isinstance(data, MultiDict):
            raise ValueError(
                "{} only deserializes {} instances".format(
                    self.__class__.__name__, MultiDict
                )
            )
        items = data.getlist(attr)
        return super()._deserialize(items, attr, data)


class RequireOnDumpMixin:
    """
    DEPRECATED AND MAY BE REMOVED IN VERSION 3.0
    In previous versions, this mixin was used to force validation on dump. As of 2.0.1, that
    validation is now fully encapsulated in compat.dump, with the presence of this mixin as one of
    the triggers.
    """

    pass


RequestSchema = Schema


class ResponseSchema(RequireOnDumpMixin, Schema):
    pass


class Error(Schema):
    message = fields.String(required=True)
    errors = fields.Dict(required=False)



from flask_rebar.utils.request_utils import *  # noqa

"""
This file is here as a temporary import-only stub/shim in order to not break any existing imports
of the following form, e.g.,:
from flask_rebar.request_utils import marshal

When we release 2.0 and can make breaking changes, this stub should be removed as we are
consolidating "utility" modules such as request_utils.py into flask_rebar.utils
"""



"""
    Test Request Utilities
    ~~~~~~~~~~~~~~~~~~~~~~

    Tests for the request utilities.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
import unittest
from tests.helpers import make_test_response

from flask import Flask
from marshmallow import fields, ValidationError

from flask_rebar import validation, response, marshal


class TestResponseFormatting(unittest.TestCase):
    def setUp(self):
        self.app = self.create_app()
        self.app.response_class = make_test_response(self.app.response_class)

    def create_app(self):
        app = Flask(__name__)

        return app

    def test_single_resource_response(self):
        @self.app.route("/single_resource")
        def handler():
            return response(data={"foo": "bar"})

        resp = self.app.test_client().get("/single_resource")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json, {"foo": "bar"})
        self.assertEqual(resp.content_type, "application/json")

    def test_single_resource_response_with_status_code(self):
        @self.app.route("/single_resource")
        def handler():
            return response(data={"foo": "bar"}, status_code=201)

        resp = self.app.test_client().get("/single_resource")
        self.assertEqual(resp.status_code, 201)
        self.assertEqual(resp.json, {"foo": "bar"})
        self.assertEqual(resp.content_type, "application/json")

    def test_single_resource_response_with_headers(self):
        header_key = "X-Foo"
        header_value = "bar"

        @self.app.route("/single_resource")
        def handler():
            return response(data={"foo": "bar"}, headers={header_key: header_value})

        resp = self.app.test_client().get("/single_resource")
        self.assertEqual(resp.headers[header_key], header_value)
        self.assertEqual(resp.json, {"foo": "bar"})
        self.assertEqual(resp.content_type, "application/json")


class SchemaForMarshaling(validation.ResponseSchema):
    foo = fields.Integer()


class TestMarshal(unittest.TestCase):
    def test_marshal(self):
        marshaled = marshal(data={"foo": 1}, schema=SchemaForMarshaling)
        self.assertEqual(marshaled, {"foo": 1})

        # Also works with an instance of the schema
        marshaled = marshal(data={"foo": 1}, schema=SchemaForMarshaling())

        self.assertEqual(marshaled, {"foo": 1})

    def test_marshal_errors(self):
        with self.assertRaises(ValidationError):
            marshal(data={"foo": "bar"}, schema=SchemaForMarshaling)
